from __future__ import annotations

import logging
import warnings
from contextlib import contextmanager
from copy import copy
from pathlib import Path
from typing import Any, Callable, Generator, Iterable, Iterator, Mapping, Optional, Sequence
from typing import Tuple
from typing import Tuple as Tup
from typing import Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colors as mcolors
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from pocketutils.core.exceptions import XValueError
from pocketutils.plotting._bound_ticks import AxisTicks, TickBounder
from pocketutils.tools.common_tools import CommonTools

logger = logging.getLogger("pocketutils")


class FigureTools:
    @classmethod
    def cm2in(cls, tup: Union[float, Iterable[float]]):
        """
        Just converts centimeters to inches.

        Args:
            tup: A float or sequence of floats (determines the return type)
        """
        if CommonTools.is_true_iterable(tup):
            return [x / 2.54 for x in tup]
        else:
            return float(tup) / 2.54

    @classmethod
    def in2cm(cls, tup: Union[float, Iterable[float]]):
        """
        Just converts inches to centimeters.

        Args:
            tup: A float or sequence of floats (determines the return type)
        """
        if CommonTools.is_true_iterable(tup):
            return [x * 2.54 for x in tup]
        else:
            return float(tup) * 2.54

    @classmethod
    def list_open_figs(cls) -> Sequence[Tuple[str, Figure]]:
        """
        Returns all currently open figures and their labels via ``Figure.label``.
        Note that ``Figure.label`` is often empty in practice.
        """
        warnings.warn("open_fig_map will be removed; use list_open_figs instead")
        return [(label, plt.figure(label=label)) for label in plt.get_figlabels()]

    @classmethod
    def open_figs(cls) -> Sequence[Figure]:
        """
        Returns all currently open figures.
        """
        warnings.warn("open_figs will be removed; use list_open_figs instead", DeprecationWarning)
        return [plt.figure(num=i) for i in plt.get_fignums()]

    @classmethod
    def open_fig_map(cls) -> Mapping[str, Figure]:
        """
        Returns all currently open figures as a dict mapping their labels ``Figure.label`` to their instances.
        Note that ``Figure.label`` is often empty in practice.
        """
        warnings.warn(
            "open_fig_map will be removed; use list_open_figs instead", DeprecationWarning
        )
        return {label: plt.figure(label=label) for label in plt.get_figlabels()}

    @classmethod
    @contextmanager
    def clearing(cls, yes: bool = True) -> Generator[None, None, None]:
        """
        Context manager to clear and close all figures created during its lifespan.
        When the context manager exits, calls ``clf`` and ``close`` on all figures created under it.

        Args:
            yes: If False, does nothing
        """
        oldfigs = copy(plt.get_fignums())
        yield
        if yes:
            for fig in [plt.figure(num=i) for i in plt.get_fignums() if i not in oldfigs]:
                fig.clf()
                plt.close(fig)

    @classmethod
    @contextmanager
    def hiding(cls, yes: bool = True) -> Generator[None, None, None]:
        """
        Context manager to hide figure display by setting ``plt.interactive(False)``.

        Args:
            yes: If False, does nothing
        """
        isint = plt.isinteractive()
        if yes:
            plt.interactive(False)
        yield
        if yes:
            plt.interactive(isint)

    @classmethod
    def plot1d(
        cls,
        values: np.array,
        *,
        figsize: Optional[Tup[float, float]] = None,
        x0=None,
        y0=None,
        x1=None,
        y1=None,
        **kwargs,
    ) -> Axes:
        """
        Plots a 1D array and returns the axes.
        kwargs are passed to ``Axes.plot``.
        """
        figure = plt.figure(figsize=figsize)
        ax = figure.add_subplot(1, 1, 1)  # Axes
        ax.plot(values, **kwargs)
        ax.set_xlim((x0, x1))
        ax.set_ylim((y0, y1))
        return ax

    @classmethod
    def despine(cls, ax: Axes) -> Axes:
        """
        Removes all spines and ticks on an Axes.
        """
        ax.set_yticks([])
        ax.set_yticks([])
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.spines["top"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])
        return ax

    @classmethod
    def clear(cls) -> int:
        """
        Removes all matplotlib figures from memory.
        Here because it's confusing to remember.
        Logs an error if not all figures were closed.

        Returns:
            The number of closed figures
        """
        n = len(plt.get_fignums())
        plt.clf()
        plt.close("all")
        m = len(plt.get_fignums())
        if m == 0:
            logger.debug(f"Cleared {n} figure{'s' if n>1 else ''}.")
        else:
            logger.error(f"Failed to close figures. Cleared {n - m}; {m} remain.")
        return n

    @classmethod
    def font_paths(cls) -> Sequence[Path]:
        """
        Returns the paths of system fonts.
        """
        # noinspection PyUnresolvedReferences
        return [Path(p) for p in matplotlib.font_manager.findSystemFonts(fontpaths=None)]

    @classmethod
    def text_matrix(
        cls,
        ax: Axes,
        data: pd.DataFrame,
        color_fn: Optional[Callable[[str], str]] = None,
        adjust_x: float = 0,
        adjust_y: float = 0,
        **kwargs,
    ) -> None:
        """
        Adds a matrix of text.

        Args:
            ax: Axes
            data: The matrix of any text values; will be converted to strings and empty strings will be ignored
            color_fn: An optional function mapping (pre-conversion-to-str) values to colors
            adjust_x: Add this value to the x coordinates
            adjust_y: Add this value to the y coordinates
            **kwargs: Passed to ``ax.text``
        """
        for r, row in enumerate(data.index):
            for c, col in enumerate(data.columns):
                value = data.iat[r, c]
                if str(value) != "":
                    ax.text(
                        r + adjust_x,
                        c + adjust_y,
                        str(value),
                        color=None if color_fn is None else color_fn(value),
                        **kwargs,
                    )

    @classmethod
    def add_note_01_coords(cls, ax: Axes, x: float, y: float, s: str, **kwargs) -> Axes:
        """
        Adds text without a box, where ``x`` and ``y`` are in coordinates (0, 1).

        Args:
            ax: axes
            x: x coord in 0-1 units
            y: y coord in 0-1 units
            s: The text
            **kwargs: Passed to ``ax.text``
        """
        t = ax.text(x, y, s=s, transform=ax.transAxes, **kwargs)
        t.set_bbox(dict(alpha=0.0))
        return ax

    @classmethod
    def add_note_data_coords(cls, ax: Axes, x: float, y: float, s: str, **kwargs) -> Axes:
        """
        Adds text without a box, where ``x`` and ``y`` are in data coordinates.

        Args:
            ax: axes
            x: x coord in data units
            y: y coord in data units
            s: The text
            **kwargs: Passed to ``ax.text``
        """
        t = ax.text(x, y, s=s, **kwargs)
        t.set_bbox(dict(alpha=0.0))
        return ax

    @classmethod
    def stamp(cls, ax: Axes, text: str, corner: str, **kwargs) -> Axes:
        """
        Adds a "stamp" in the corner.

        Example:
            Stamping::

                FigureTools.stamp(ax, 'hello', "top_right")

        Args:
            ax: Axes
            text: The text
            corner: Ignoring ``"_"``, ``"-"``, ``" "``, and case,
                    must be "topleft", "topright", "bottomleft", or "bottomright".
            **kwargs: Passed to ``ax.text``;
                      do not pass "x", "y", "horizontalalignment", or "verticalalignment"
        """
        return cls._text(ax, text, corner, **kwargs)

    @classmethod
    def alignment_for_corner(cls, corner: str) -> Mapping[str, Any]:
        """
        Get matplotlib arguments appropriate for text in a corner.

        Args:
            corner: Ignoring ``"_"``, ``"-"``, ``" "``, and case,
                    must be "topleft", "topright", "bottomleft", or "bottomright".
        """
        corner = corner.lower().replace(" ", "").replace("-", "").replace("_", "")
        valid = {"topleft", "topright", "bottomleft", "bottomright"}
        if corner not in valid:
            raise XValueError(f"Corner {corner} not in {valid} (_, ' ', and - ignored)")
        left = corner.endswith("left")
        top = corner.startswith("top")
        return dict(
            x=0 if left else 1,
            y=1 if top else 0,
            horizontalalignment="left" if left else 1,
            verticalalignment="top" if top else "bottom",
        )

    @classmethod
    def plot_palette(cls, values: Union[Sequence[str], str]) -> Figure:
        """
        Plots a color palette.

        Args:
            values: A string of a color (starting with #), a sequence of colors (each starting with #)
        """
        n = len(values)
        figure = plt.figure(figsize=(8.0, 2.0))
        ax = figure.add_subplot(1, 1, 1)
        ax.imshow(
            np.arange(n).reshape(1, n),
            cmap=mcolors.ListedColormap(values),
            interpolation="none",
            aspect="auto",
        )
        cls.despine(ax)
        return figure

    @classmethod
    def bind_to_ticks(
        cls,
        ax: Axes,
        *,
        x0: Union[bool, float] = True,
        y0: Union[bool, float] = True,
        x1: Union[bool, float] = True,
        y1: Union[bool, float] = True,
        major: bool = True,
    ) -> Axes:
        """
        Forces Axes limits to start and/or end at major or minor ticks.

        Each argument in the constructor can be:
        - ``True`` -- set the bound according to the ticks
        - ``False`` -- do not change the bound
        - a ``float`` - set to this

        Args:
            ax: The axes, of course
            x0: left bound (``.getxlim()[0]``)
            y0: bottom bound (``.getylim()[0]``)
            x1: right bound (``.getxlim()[1]``)
            y1: top bound (``.getylim()[1]``)
            major: Use major tick marks rather than minor
        """
        return TickBounder(x0=x0, y0=y0, x1=x1, y1=y1, major=major).adjust(ax)

    @classmethod
    def _text(cls, ax: Axes, text: str, corner: str, **kwargs) -> Axes:
        t = ax.text(s=text, **cls.alignment_for_corner(corner), transform=ax.transAxes, **kwargs)
        t.set_bbox(dict(alpha=0.0))
        return ax


__all__ = ["FigureTools"]



import logging
import os
from typing import Iterable, List, Mapping, Optional, Union
from urllib import request

import pandas as pd
import regex
import uniprot

# uses https://github.com/tanghaibao/goatools
from goatools import obo_parser

# NOT the same as FlatGoTerm, which has no knowledge of hierarchy
from goatools.obo_parser import GOTerm

from pocketutils.core.exceptions import MultipleMatchesError, StringPatternError

# noinspection PyProtectedMember
from pocketutils.core.input_output import silenced

go_pattern = regex.compile(
    r"GO:(\d+); ([CFP]):([\dA-Za-z- ,()]+); ([A-Z]+):([A-Za-z-_]+)\.", flags=regex.V1
)
logger = logging.getLogger("pocketutils")


class FlatGoTerm:
    """
    A Gene Ontology term.
    Not to be confused with GOTerm in goatools: obo_parser.GOTerm

    Attributes:
        ID (str); ex: GO:0005737
        kind (str: 'P'==process, 'C'==component, 'F'==function)
        description (str)
        sourceId (str); ex: IDA
        sourceName (str); ex: UniProtKB
    """

    def __init_(
        self,
        identifier: str,
        kind: str,
        description: str,
        source_id: str,
        source_name: str,
    ):
        self.ID = identifier
        self.kind = kind
        self.description = description
        self.sourceId = source_id
        self.sourceName = source_name

    def __init__(self, stwing: str):
        """Builds a GO term from a string from uniprot_obj['go'].
        Raises a ValueError if the syntax is wrong.
        """
        match = go_pattern.search(stwing)
        if match is None:
            raise StringPatternError(
                f"String didn't match GO term pattern: {stwing}",
                value=stwing,
                pattern=go_pattern,
            )
        self.ID = "GO:" + match.group(1)
        self.kind = match.group(2)
        self.description = match.group(3)
        self.sourceId = match.group(4)
        self.sourceName = match.group(5)

    def to_series(self) -> pd.Series:
        return pd.Series(self.__dict__)


class UniprotGoTerms:
    def fetch_uniprot_data(self, uniprot_ids: Union[str, List[str]]) -> List[Mapping[str, str]]:
        """
        Fetches a list of dicts of UniProt metadata, one per UniProt ID.

        Raises:
            ValueError: If a UniProt ID wasn't found.
        """
        if isinstance(uniprot_ids, str):  # not a list type
            uniprot_ids = [uniprot_ids]
        # if we don't prevent these here, we'll get a ValueError from below, which is confusing
        # That's because uniprot.fetch_uniprot_metadata will only return one per unique ID
        if len(set(uniprot_ids)) != len(uniprot_ids):
            raise MultipleMatchesError("Set of UniProt IDs cannot contain duplicates")
        with silenced(no_stderr=False):
            uniprot_data = uniprot.fetch_uniprot_metadata(uniprot_ids)
        if uniprot_data is None or uniprot_data == {} or len(uniprot_data) != len(uniprot_ids):
            raise LookupError(f"At least one UniProt ID not found in {uniprot_ids}")
        return list(uniprot_data.values())

    def go_terms_for_uniprot_id(self, uniprot_id: str) -> List[FlatGoTerm]:
        """Returns a list of FlatGoTerm objects from a UniProt ID."""
        term_strings = (self.fetch_uniprot_data(uniprot_id)[0])["go"]
        return [FlatGoTerm(s) for s in term_strings]

    def go_terms_for_uniprot_id_as_df(self, uniprot_id: str) -> pd.DataFrame:
        """Returns a Pandas DataFrame of GO terms from a UniProt ID."""
        df = pd.DataFrame(columns=["ID", "kind", "description", "sourceId", "sourceName"])
        for term in self.go_terms_for_uniprot_id(uniprot_id):
            df.loc[len(df)] = term.to_series()
        return df.set_index("ID")


class GoTermsAtLevel:
    """
    Gene ontology terms organized by level.
    Example:
        >>> go_term_ancestors_for_uniprot_id_as_df('P42681', 2)
    """

    def __init__(self) -> None:
        if os.path.exists("gene_ontology.1_2.obo"):
            self.obo = obo_parser.GODag("gene_ontology.1_2.obo")
        else:
            logger.info("Downloading Gene Ontology OBO...")
            request.urlretrieve(
                "http://www.geneontology.org/ontology/obo_format_1_2/gene_ontology.1_2.obo"
            )
            self.obo = obo_parser.GODag(
                "gene_ontology.1_2.obo"
            )  # This will be used in query_obo_term
            logger.info("Done downloading OBO.")
        self.substruct = UniprotGoTerms()

    def query_obo_term(self, term_id: str) -> GOTerm:
        """Queries a term through the global obo.
        This function wraps the call to raise a ValueError if the term is not found;
        otherwise it only logs a warning.
        """
        x = self.obo.query_term(term_id)
        if x is None:
            raise LookupError(f"Term ID {x} not found")
        return x

    def get_ancestors_of_go_term(self, term_id: str, level: int) -> Iterable[GOTerm]:
        """
        From a GO term in the form 'GO:0007344', returns a set of ancestor GOTerm objects at the specified level.
        The traversal is restricted to is-a relationships.
        Note that the level is the minimum number of steps to the root.

        Args:
            level: starting at 0 (root)
        """

        def traverse_up(term, buildup_set, level):
            if term.level == level:
                buildup_set.add(term)
            if term.has_parent:
                return [traverse_up(p, buildup_set, level) for p in term.parents]
            return None

        terms = set()
        traverse_up(self.query_obo_term(term_id), terms, level)
        return terms

    def go_term_ancestors_for_uniprot_id(
        self, uniprot_id: str, level: int, kinds_allowed: Optional[List[str]] = None
    ) -> Iterable[GOTerm]:
        """
        Gets the GO terms associated with a UniProt ID and returns a set of their ancestors at the specified level.
        The traversal is restricted to is-a relationships.
        Note that the level is the minimum number of steps to the root.

        Args:
            level: starting at 0 (root)
            kinds_allowed: a set containing any combination of 'P', 'F', or 'C'
        """
        if kinds_allowed is None:
            kinds_allowed = ["P", "F", "C"]
        if len(kinds_allowed) == 0:
            return []
        terms = [
            term
            for term in self.substruct.go_terms_for_uniprot_id(uniprot_id)
            if term.kind in kinds_allowed
        ]
        ancestor_terms = set()
        for term_id in [t.ID for t in terms]:
            ancestor_terms.update(self.get_ancestors_of_go_term(term_id, level))
        return ancestor_terms

    def go_term_ancestors_for_uniprot_id_as_df(
        self, uniprot_id: str, level: int, kinds_allowed: Optional[List[str]] = None
    ) -> pd.DataFrame:
        if kinds_allowed is None:
            kinds_allowed = ["P", "F", "C"]
        """See go_term_ancestors_for_uniprot_id. Returns a Pandas DataFrame with columns IDand name."""
        df = pd.DataFrame(columns=["ID", "name"])
        for term in self.go_term_ancestors_for_uniprot_id(uniprot_id, level, kinds_allowed):
            df.loc[len(df)] = pd.Series({"ID": term.id, "name": term.name, "level": term.level})
        return df.set_index("ID")


__all__ = ["FlatGoTerm", "UniprotGoTerms", "GoTermsAtLevel"]



import enum
import logging
from typing import AbstractSet, Optional, Union

from pocketutils.core.exceptions import XValueError

logger = logging.getLogger("pocketutils")


class DisjointEnum(enum.Enum):
    """
    An enum that does not have combinations.
    """

    @classmethod
    def _fix_lookup(cls, s: str) -> str:
        return s

    @classmethod
    def or_none(cls, s: Union[str, __qualname__]) -> Optional[__qualname__]:
        """
        Returns a choice by name (or returns ``s`` itself).
        Returns ``None`` if the choice is not found.
        """
        try:
            return cls.of(s)
        except KeyError:
            return None

    @classmethod
    def of(cls, s: Union[str, __qualname__]) -> __qualname__:
        """
        Returns a choice by name (or returns ``s`` itself).
        """
        if isinstance(s, cls):
            return s
        return cls[cls._fix_lookup(s)]

    def __new__(cls, *args, **kwargs):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class FlagEnum(enum.Flag):
    """
    A bit flag that behaves as a set, has a null set, and auto-sets values and names.

    Example:
        .. code-block::

            class Flavor(FlagEnum):
                none = ()
                bitter = ()
                sweet = ()
                sour = ()
                umami = ()
            bittersweet = Flavor.bitter | Flavor.sweet
            print(bittersweet.value)  # 1 + 2 == 3
            print(bittersweet.name)   # "bitter|sweet"

    .. important::
        The *first element* must always be the null set ("no flags")
        and should be named something like 'none', 'empty', or 'zero'
    """

    @classmethod
    def _fix_lookup(cls, s: str) -> str:
        return s

    def __new__(cls, *args, **kwargs):
        if len(cls.__members__) == 0:
            value = 0
        else:
            value = 2 ** (len(cls.__members__) - 1)
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    @classmethod
    def _create_pseudo_member_(cls, value):
        value = super()._create_pseudo_member_(value)
        members, _ = enum._decompose(cls, value)
        value._name_ = "|".join([m.name for m in members])
        return value

    @classmethod
    def or_none(cls, s: Union[str, __qualname__]) -> Optional[__qualname__]:
        """
        Returns a choice by name (or returns ``s`` itself).
        Returns ``None`` if the choice is not found.
        """
        try:
            return cls.of(s)
        except KeyError:
            return None

    @classmethod
    def of(cls, s: Union[str, __qualname__, AbstractSet[Union[str, __qualname__]]]) -> __qualname__:
        """
        Returns a choice by name (or ``s`` itself), or a set of those.
        """
        if isinstance(s, cls):
            return s
        if isinstance(s, str):
            return cls[cls._fix_lookup_(s)]
        z = cls[0]
        for m in s:
            z |= cls.of(m)
        return z

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name


class TrueFalseUnknown(DisjointEnum):
    """
    A :class:`pocketutils.core.enums.DisjointEnum` of true, false, or unknown.
    """

    true = ()
    false = ()
    unknown = ()

    @classmethod
    def _unmatched_type(cls) -> Optional[__qualname__]:
        return cls.unknown

    @classmethod
    def _fix_lookup(cls, s: str) -> str:
        s = s.lower().strip()
        return dict(t="true", false="false").get(s, s)


class MultiTruth(FlagEnum):
    """
    A :class:`pocketutils.core.enums.FlagEnum` for true, false, true+false, and neither.
    """

    false = ()
    true = ()


class CleverEnum(DisjointEnum):
    """
    An enum with a ``.of`` method that finds values with limited string/value fixing.
    Replaces ``" "``, ``"-"``, and ``"."`` with ``_`` and ignores case in :meth:`of`.
    May support an "unmatched" type -- a fallback value when there is no match.
    This is similar to the simpler :class:`pocketutils.core.SmartEnum`.
    """

    @classmethod
    def of(cls, s: Union[str, __qualname__]) -> __qualname__:
        try:
            return super().of(s)
        except KeyError:
            unknown = cls._unmatched_type()
            logger.error(f"Value {s} not found. Using {unknown}")
            if unknown is None:
                raise XValueError(f"Value {s} not found and unmatched_type is None")
            return unknown

    @classmethod
    def _unmatched_type(cls) -> Optional[__qualname__]:
        return None

    @classmethod
    def _fix_lookup(cls, s: str) -> str:
        return s.strip().replace(" ", "_").replace(".", "_").replace("-", "_").lower()


__all__ = ["TrueFalseUnknown", "DisjointEnum", "FlagEnum", "CleverEnum"]



from __future__ import annotations

import logging
from copy import copy
from pathlib import Path
from typing import Any, Callable, Generator, Iterator, Mapping, Optional, Sequence
from typing import Tuple as Tup
from typing import Union
from warnings import warn

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.figure import Figure

from pocketutils.core import PathLike

# these have no extra dependencies
from pocketutils.core.exceptions import IllegalPathError, XFileExistsError, XValueError
from pocketutils.tools.common_tools import CommonTools
from pocketutils.tools.path_tools import PathTools

FigureSeqLike = Union[Figure, Iterator[Figure], Iterator[Tup[str, Figure]], Mapping[str, Figure]]
KNOWN_EXTENSIONS = ["jpg", "png", "pdf", "svg", "eps", "ps"]
logger = logging.getLogger("pocketutils")


class FigureSaver:
    """
    Offers some small, specific extensions over matplotlib's ``figure.savefig``:
    - can remove the figure from memory each iteration
    - creates directories as needed
    - can save a multi-figure PDF
    - complains about issues
    - can auto-fix some
    - the ``FigureSaver.save`` function handles iterators of different types

    See ``FigureSaver.save`` for more info.
    ``clear`` defines the behavior after saving a single figure:
        - False    ==> do nothing
        - True     ==> clear it
        - callable ==> call it with the Figure instance

    """

    def __init__(
        self,
        save_under: Optional[PathLike] = None,
        *,
        clear: Union[bool, Callable[[Figure], Any]] = False,
        as_type: Optional[str] = None,
        check_paths: bool = True,
        sanitize_paths: bool = True,
        log: Optional[Callable[[str], None]] = None,
        kwargs: Mapping[str, Any] = None,
    ):
        """
        Constructor.

        Args:
            save_under: A parent directory assumed for all saved figures
            clear: Auto-close each figure after saving
            as_type: Force a specific filename suffix (e.g. ``as_type=".jpg")``
            check_paths: Check that paths are valid; without ``sanitize_paths``,
                         raises an error if a path is invalid
            sanitize_paths: Try to sanitize paths (implies ``check_paths``)
            log: Call this function with a message indicating that a sanitized path differs
            kwargs: Passed to ``savefig``
        """
        self._save_under = None if save_under is None else Path(save_under)
        if clear is not None and not isinstance(clear, bool) and not callable(clear):
            raise TypeError(type(clear))
        self._clear = clear
        self._log = lambda _: None if log is None else log
        self._as_type = "." + as_type.lstrip(".")
        self._check_paths = check_paths
        self._sanitize_paths = sanitize_paths
        self._kwargs = {} if kwargs is None else kwargs

    def __mod__(self, tup):
        """
        See: :meth:`save_one`.
        """
        self.save_one(*tup)

    def __idiv__(self, tup):
        """
        See: :meth:`save`.
        """
        self.save(*tup)

    def save(
        self,
        figure: FigureSeqLike,
        path: PathLike = "",
        names: Optional[Iterator[str]] = None,
        *,
        overwrite: bool = True,
        use_labels: bool = False,
    ) -> None:
        """
        Saves either:
            1. a single figure ``figure`` to path ``path``
            2. a bunch of figures to directory ``path``
               if ``figure`` is an iterable (list, dict, etc) over figures
            3. a single PDF with multiple figures, if ``path`` ends in ``.pdf``

        If ``figure`` is iterable (case 2), it can be either:
            - an iterable over Figures
            - an iterable over (name, Figure) pairs, where ``name`` provides the filename
              (under the directory ``path``)

        If it's the first case and ``names`` is set, will use those to provide the filenames.
        Otherwise, falls back to numbering them (ex: ``directory/1.png``, etc).

        If ``use_labels`` is set, tries to use each ``Figure.get_label()``.

        .. note ::
            Figure labels are not typically set by default;
            you should call ``Figure.set_label()`` when using ``use_labels``.
        """
        is_iterable = CommonTools.is_true_iterable(figure)
        if is_iterable and path.endswith(".pdf"):
            self.save_all_as_pdf(figure, path, names=names)
        elif is_iterable:
            self.save_all(figure, path, names=names, use_labels=use_labels, overwrite=overwrite)
        else:
            self.save_one(figure, path, overwrite=overwrite)

    def save_all_as_pdf(
        self,
        figures: FigureSeqLike,
        path: PathLike,
        names: Optional[Iterator[str]] = None,
        *,
        overwrite: bool = True,
        metadata: Optional[Mapping[str, str]] = None,
    ) -> None:
        """
        Save a single PDF with potentially many figures.
        See :meth:`save` for more info.
        """
        # note! this is weird
        if plt.rcParams["pdf.fonttype"] != 42:
            warn("pdf.fonttype != 42. This may cause problems.")
        cp = copy(self)
        cp._as_type = ".pdf"
        path = cp._sanitized_file(path, overwrite=overwrite)
        done_figures = []
        with PdfPages(str(path), keep_empty=False, metadata=metadata) as pdf:
            for name, figure in self._enumerate(figures, names, use_labels=False):
                pdf.savefig(figure, **self._kwargs)
                done_figures.append(figure)
        # TODO can I clear every time?
        for figure in done_figures:
            self._clean_up(figure)

    def save_all(
        self,
        figures: FigureSeqLike,
        directory: PathLike = ".",
        names: Optional[Iterator[str]] = None,
        *,
        use_labels: bool = False,
        overwrite: bool = True,
    ) -> None:
        """
        Saves potentially multiple figures into a directory.

        See :meth:`save` for more info.
        """
        for name, figure in self._enumerate(figures, names, use_labels=use_labels):
            # DO NOT prepend self.__save_under here!! It's done in save_one.
            path = Path(directory) / name
            self.save_one(figure, path, overwrite=overwrite)  # clears if needed

    def save_one(self, figure: Figure, path: PathLike, overwrite: bool) -> None:
        """
        Saves one figure to a specific path.
        This is pretty similar to ``Figure.savefig``.
        """
        path = self._sanitized_file(path, overwrite=overwrite)
        figure.savefig(path, **self._kwargs)
        self._clean_up(figure)

    def _enumerate(
        self,
        figures,
        names: Optional[Sequence[str]],
        *,
        use_labels: bool,
    ) -> Generator[Tup[str, Figure], None, None]:
        if isinstance(figures, Mapping):
            figures = figures.items()
        for i, figure in enumerate(figures):
            if names is not None:
                yield next(names), figure
            elif isinstance(figure, tuple):
                yield figure[0], figure[1]
            elif use_labels and figure.get_label() is not None:
                yield figure.get_label(), figure
            elif use_labels:
                raise ValueError("use_labels=True, but no names given and a figure has no label")
            else:
                yield str(i), figure

    def _clean_up(self, figure: Figure) -> None:
        if self._clear is None:
            pass
        elif self._clear is True:
            pass
            figure.clear()
            figure.clf()
        elif callable(self._clear):
            self._clear(figure)

    def _sanitized_file(self, path: PathLike, overwrite: bool) -> Path:
        """
        Sanitizes a file path:
            - prepends self._save_under if needed
            - warns about issues

        Args:
            path: The path, including directory, but excluding self._save_under
        """
        path = Path(path)
        if self._check_paths or self._sanitize_paths:
            new_path = PathTools.sanitize_path(path, show_warnings=False)
            if new_path != path and self._sanitize_paths:
                self._log(f"Sanitized filename {path} → {new_path}")
            elif new_path != path:
                raise IllegalPathError(
                    f"Filename {path} is not valid (could be: {new_path})", path=path
                )
        if self._save_under is not None and Path(path).is_absolute():
            raise XValueError(f"_save_under is {self._save_under} but path {path} is absolute")
        elif self._save_under is not None:
            path = self._save_under / path
        ext_valid = any((str(path).endswith("." + s) for s in KNOWN_EXTENSIONS))
        # can't just detect no suffix, because 12.5uM will have suffix 5uM
        # also, don't use with_suffix for this same reason
        if not ext_valid:
            pt = "." + plt.rcParams["savefig.format"] if self._as_type is None else self._as_type
            path = Path(str(path) + pt)
        path = Path(path)
        if path.exists() and not overwrite:
            raise XFileExistsError(f"{path} already exists", path=path)
        path.parent.mkdir(exist_ok=True, parents=True)
        logger.debug(path)
        return path


__all__ = ["FigureSaver"]



import logging
import sys
from contextlib import contextmanager
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from pocketutils.core._internal import look as _look
from pocketutils.core.exceptions import LengthError, LengthMismatchError, MultipleMatchesError

logger = logging.getLogger("pocketutils")
Y = TypeVar("Y")
T = TypeVar("T")
Z = TypeVar("Z")


class BaseTools:
    @classmethod
    def is_lambda(cls, function: Any) -> bool:
        """
        Returns whether this is a lambda function. Will return False for non-callables.
        """
        LAMBDA = lambda: 0
        if not hasattr(function, "__name__"):
            return False  # not a function
        return (
            isinstance(function, type(LAMBDA))
            and function.__name__ == LAMBDA.__name__
            or str(function).startswith("<function <lambda> at ")
            and str(function).endswith(">")
        )

    @classmethod
    def only(
        cls,
        sequence: Iterable[Any],
        condition: Union[str, Callable[[Any], bool]] = None,
        name: str = "collection",
    ) -> Any:
        """
        Returns either the SINGLE (ONLY) UNIQUE ITEM in the sequence or raises an exception.
        Each item must have __hash__ defined on it.

        Args:
            sequence: A list of any items (untyped)
            condition: If nonnull, consider only those matching this condition
            name: Just a name for the collection to use in an error message

        Returns:
            The first item the sequence.

        Raises:
            LookupError If the sequence is empty
            MultipleMatchesError If there is more than one unique item.
        """

        def _only(sq):
            st = set(sq)
            if len(st) > 1:
                raise MultipleMatchesError("More then 1 item in " + str(name))
            if len(st) == 0:
                raise LookupError("Empty " + str(name))
            return next(iter(st))

        if condition and isinstance(condition, str):
            return _only(
                [
                    s
                    for s in sequence
                    if (
                        not getattr(s, condition[1:])
                        if condition.startswith("!")
                        else getattr(s, condition)
                    )
                ]
            )
        elif condition:
            return _only([s for s in sequence if condition(s)])
        else:
            return _only(sequence)

    @classmethod
    def zip_strict(cls, *args: Iterable[Any]) -> Generator[Tuple[Any], None, None]:
        """
        Same as zip(), but raises an IndexError if the lengths don't match.

        Args:
            args: Same as with ``zip``

        Yields:
            Tuples corresponding to the input items

        Raises:
            LengthMismatchError: If the lengths of the input iterators don't match
        """
        # we need to catch these cases before or they'll fail
        # in particular, 1 element would fail with a LengthMismatchError
        # and 0 elements would loop forever
        if len(args) < 2:
            yield from zip(*args)
            return
        iters = [iter(axis) for axis in args]
        n_elements = 0
        failures = []
        while len(failures) == 0:
            values = []
            failures = []
            for axis, iterator in enumerate(iters):
                try:
                    values.append(next(iterator))
                except StopIteration:
                    failures.append(axis)
            if len(failures) == 0:
                yield tuple(values)
            elif len(failures) == 1:
                raise LengthError(f"Too few elements ({n_elements}) along axis {failures[0]}")
            elif len(failures) < len(iters):
                raise LengthError(f"Too few elements ({n_elements}) along axes {failures}")
            n_elements += 1

    @classmethod
    def zip_list(cls, *args) -> List[Tuple[Any]]:
        """
        Same as ``zip_strict``, but more informative.
        Converts to a list and can provide a more detailed error message.
        Zips two sequences into a list of tuples and raises an
        IndexError if the lengths don't match.

        Args:
            args: Same as with ``zip``

        Yields:
            Tuples corresponding to the input items

        Raises:
            LengthMismatchError: If the lengths of the input iterators don't match
        """
        try:
            return list(cls.zip_strict(*args))
        except LengthMismatchError:
            raise LengthMismatchError(
                "Length mismatch in zip_strict: Sizes are {}".format([len(x) for x in args])
            ) from None

    @classmethod
    def forever(cls) -> Iterator[int]:
        """
        Yields i for i in range(0, infinity).
        Useful for simplifying a i = 0; while True: i += 1 block.
        """
        i = 0
        while True:
            yield i
            i += 1

    @classmethod
    def to_true_iterable(cls, s: Any) -> Iterable[Any]:
        """
        See :meth:`is_true_iterable`.

        Examples:
            - ``to_true_iterable('abc')         # ['abc']``
            - ``to_true_iterable(['ab', 'cd')]  # ['ab', 'cd']``
        """
        if BaseTools.is_true_iterable(s):
            return s
        else:
            return [s]

    @classmethod
    def is_true_iterable(cls, s: Any) -> bool:
        """
        Returns whether ``s`` is a probably "proper" iterable.
        In other words, iterable but not a string or bytes
        """
        return (
            s is not None
            and isinstance(s, Iterable)
            and not isinstance(s, str)
            and not isinstance(s, bytes)
        )

    @classmethod
    @contextmanager
    def null_context(cls) -> Generator[None, None, None]:
        """
        Returns an empty context (literally just yields).
        Useful to simplify when a generator needs to be used depending on a switch.
        Ex:
            if verbose_flag:
                do_something()
            else:
                with Tools.silenced():
                    do_something()
        Can become:
            with (Tools.null_context() if verbose else Tools.silenced()):
                do_something()
        """
        yield

    @classmethod
    def look(cls, obj: Y, attrs: Union[str, Iterable[str], Callable[[Y], Z]]) -> Optional[Z]:
        """
        Returns the value of a chain of attributes on object ``obj``,
        or None any object in that chain is None or lacks the next attribute.

        Example:
            Get a kitten's breed::

            BaseTools.look(kitten), 'breed.name')  # either None or a string

        Args:
            obj: Any object
            attrs: One of:
                - A string in the form attr1.attr2, translating to ``obj.attr1``
                - An iterable of strings of the attributes
                - A function that maps ``obj`` to its output;
                   equivalent to calling `attrs(obj)` but returning None on ``AttributeError``.

        Returns:
            Either None or the type of the attribute

        Raises:
            TypeError:
        """
        return _look(obj, attrs)

    @classmethod
    def get_log_function(
        cls, log: Union[None, str, Callable[[str], None], Any]
    ) -> Callable[[str], None]:
        """
        Gets a logging function from user input.
        The rules are:
            - If None, uses logger.info
            - If 'print' or 'stdout',  use sys.stdout.write
            - If 'stderr', use sys.stderr.write
            - If another str or int, try using that logger level (raises an error if invalid)
            - If callable, returns it
            - If it has a callable method called 'write', uses that

        Returns:
            A function of the log message that returns None
        """
        if log is None:
            return logger.info
        elif isinstance(log, str) and log.lower() in ["print", "stdout"]:
            # noinspection PyTypeChecker
            return sys.stdout.write
        elif log == "stderr":
            # noinspection PyTypeChecker
            return sys.stderr.write
        elif isinstance(log, int):
            return getattr(logger, logging.getLevelName(log).lower())
        elif isinstance(log, str):
            return getattr(logger, log.lower())
        elif callable(log):
            return log
        elif hasattr(log, "write") and getattr(log, "write"):
            return getattr(log, "write")
        else:
            raise TypeError(f"Log type {type(log)} not known")

    def __repr__(self):
        return self.__class__.__name__

    def __str__(self):
        return self.__class__.__name__


__all__ = ["BaseTools"]



from __future__ import annotations

import hashlib
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional, Union

from pocketutils.core.exceptions import (
    FileDoesNotExistError,
    HashValidationFailedError,
    IllegalStateError,
)


@dataclass(frozen=True, unsafe_hash=True, repr=True)
class HashableFile:
    """
    A file ``path`` with an associated hash file named ``path.with_suffix(path.suffix + "." + suffix)``.
    For example, the path might be ``x.tar.gz`` and the hash file ``x.tar.gz.sha1``.

    There are three valid states:
        - non-hashed (HashableFile)
        - post-hashed (PostHashedFile)
        - pre-hashed (PreHashedFile)

        file_path: The path of the file
        hash_path: The path of the hash
    """

    file_path: Path
    hash_path: Path
    actual: Optional[str]
    expected: Optional[str]
    algorithm: Callable[[], Any]
    buffer_size: int = 16 * 1024

    def __post_init__(self):
        warnings.warn("hashers will be removed; use typeddfs's Checksums instead")
        if not self.file_path.exists():
            raise FileDoesNotExistError(f"File {self.file_path} not found", path=self.file_path)

    @property
    def files_exist(self) -> bool:
        return self.file_path.exists() and self.hash_path.exists()

    def computed(self) -> PostHashedFile:
        return PostHashedFile(
            file_path=self.file_path,
            hash_path=self.hash_path,
            actual=self._get_or_compute(),
            expected=self.expected,
            algorithm=self.algorithm,
            buffer_size=self.buffer_size,
        )

    def precomputed(self) -> PreHashedFile:
        if not self.hash_path.exists():
            raise IllegalStateError(f"Hash file {self.hash_path} does not exist")
        return PreHashedFile(
            file_path=self.file_path,
            hash_path=self.hash_path,
            expected=self._read_hash(),
            actual=self.actual,
            algorithm=self.algorithm,
            buffer_size=self.buffer_size,
        )

    def compute(self) -> str:
        alg = self.algorithm()
        with self.file_path.open("rb") as f:
            for chunk in iter(lambda: f.read(self.buffer_size), b""):
                alg.update(chunk)
        return alg.hexdigest()

    def _read_hash(self) -> str:
        return self.hash_path.read_text(encoding="utf8").strip()

    def _write_hash(self) -> None:
        self.hash_path.write_text(self.actual, encoding="utf8")

    def _get_or_compute(self) -> str:
        if self.actual is not None:
            return self.actual
        return self.compute()


@dataclass(frozen=True, unsafe_hash=True, repr=True)
class NonHashedFile(HashableFile):
    def __post_init__(self):
        super().__post_init__()
        if self.hash_path.exists():
            raise IllegalStateError(f"Hash file {self.hash_path} already exists")


@dataclass(frozen=True, unsafe_hash=True, repr=True)
class PostHashedFile(HashableFile):
    def precomputed(self) -> PrePostHashedFile:
        if not self.hash_path.exists():
            raise IllegalStateError(f"Hash file {self.hash_path} does not exist")
        return PrePostHashedFile(
            file_path=self.file_path,
            hash_path=self.hash_path,
            expected=self._read_hash(),
            actual=self.actual,
            algorithm=self.algorithm,
            buffer_size=self.buffer_size,
        )

    def write(self) -> None:
        """
        Writes the actual (computed) hash to the file.
        Does not affect the state, except that the file will exist.

        Raises:
            IllegalStateError: If the hash file already exists.
                               This can only be true if it was written after instantiating this.
        """
        if self.hash_path.exists():
            raise IllegalStateError(f"Hash file {self.hash_path} already exists")
        self._write_hash()

    def __post_init__(self):
        super().__post_init__()
        if self.hash_path.exists():
            raise IllegalStateError(f"Hash file {self.hash_path} already exists")
        if self.actual is None:
            raise IllegalStateError(f"Actual hash value does not exist for path {self.file_path}")


@dataclass(frozen=True, unsafe_hash=True, repr=True)
class PreHashedFile(HashableFile):
    def __post_init__(self):
        super().__post_init__()
        if self.expected is None:
            raise IllegalStateError(f"Expected hash value does not exist for path {self.file_path}")

    def computed(self) -> PrePostHashedFile:
        return PrePostHashedFile(
            file_path=self.file_path,
            hash_path=self.hash_path,
            actual=self._get_or_compute(),
            expected=self.expected,
            algorithm=self.algorithm,
            buffer_size=self.buffer_size,
        )


@dataclass(frozen=True, unsafe_hash=True, repr=True)
class PrePostHashedFile(HashableFile):
    def __post_init__(self):
        super().__post_init__()
        if self.expected is None:
            raise ValueError(f"Expected hash value does not exist for path {self.file_path}")
        if self.actual is None:
            raise ValueError(f"Actual hash value does not exist for path {self.file_path}")

    def match_or_raise(self) -> str:
        if not self.matches:
            raise HashValidationFailedError(
                f"Hash for file {self.file_path} does not match",
                key=self.file_path,
                expected=self.expected,
                actual=self.actual,
            )
        return self.expected

    @property
    def matches(self) -> bool:
        return self.actual == self.expected


class Hasher:
    """
    Makes and reads .sha1 / .sha256 files next to existing paths.
    """

    def __init__(
        self,
        algorithm: str = "sha1",
        buffer_size: int = 16 * 1024,
    ):
        self._algorithm = algorithm
        if isinstance(algorithm, str):
            self._algorithm_class = getattr(hashlib, algorithm)
            self._suffix = algorithm
        self._buffer_size = buffer_size

    @property
    def algorithm(self) -> str:
        return self._algorithm

    def to_write(self, path: Union[Path, str]) -> PostHashedFile:
        """
        Gets a HashableFile that has an already-existing
        Computes the actual hash from the file.
        Complains if the hash file already exists.
        """
        return self._new(path).computed()

    def to_verify(self, path: Union[Path, str]) -> PrePostHashedFile:
        """
        Gets a HashableFile that has an existing hash file.
        Computes the actual hash from the file.
        Complains if the hash file does not exist.
        """
        return self._new(path).precomputed().computed()

    def any(
        self, path: Union[Path, str], computed: bool = False, precomputed: Optional[bool] = False
    ) -> HashableFile:
        new = self._new(path)
        if computed and (precomputed or (precomputed is None and new.hash_path.exists())):
            return new.precomputed().computed()
        elif precomputed or (precomputed is None and new.hash_path.exists()):
            return new.precomputed()
        elif computed:
            return new.computed()
        else:
            return new

    def _new(self, path: Path):
        path = Path(path)
        hash_path = path.with_suffix(path.suffix + "." + self._suffix)
        return HashableFile(
            file_path=path,
            hash_path=hash_path,
            expected=None,
            actual=None,
            algorithm=self._algorithm_class,
            buffer_size=self._buffer_size,
        )


__all__ = [
    "Hasher",
    "HashableFile",
    "NonHashedFile",
    "PreHashedFile",
    "PostHashedFile",
    "PrePostHashedFile",
]



from typing import Any


class Chars:
    """Unicode symbols that are useful in code and annoying to search for repeatedly."""

    # punctuation
    nbsp = "\u00A0"  # non-breaking space
    zwidthspace = "\u200B"  # zero-width space
    thinspace = "\u2009"
    hairspace = "\u200A"
    emspace = "\u2003"
    figspace = "\u2007"
    narrownbsp = "\u202F"  # great for units
    hyphen = "‐"  # proper unicode hyphen
    nbhyphen = "‑"  # non-breaking hyphen
    fig = "‒"  # figure dash, ex in phone numbers
    en = "–"  # en dash, ex in ranges
    em = "—"  # em dash, like a semicolon
    ellipsis = "…"  # only 1 character, which is helpful
    middots = "⋯"
    middot = "·"
    rsq, lsq, rdq, ldq = "’", "‘", "”", "“"
    # math
    ell = "ℓ"
    micro, degree, angstrom = "µ", "°", "Å"
    minus, times, plusminus = "−", "×", "±"
    inf, null = "∞", "⌀"
    prop, approx, leq, geq = "∝", "≈", "≤", "≥"
    nott, implies, iff, forall, exists, notexists = "¬", "⇒", "⇔", "∀", "∃", "∄"
    vee, wedge, cup, cap = "∨", "∧", "∪", "∩"
    isin, contains, complement = "∈", "∋", "∁"
    precedes, succeeds = "≺", "≻"
    prime, partial, integral = "′", "∂", "∫"
    # info marks
    bullet = "•"
    dagger, ddagger = "†", "‡"
    star, snowflake = "★", "⁕"
    info, caution, warning, donotenter, noentry = "🛈", "☡", "⚠", "⛔", "🚫"
    trash, skull, atom, radiation, bioharzard = "🗑", "☠", "⚛", "☢", "☣"
    corners = "⛶"
    # misc / UI
    left, right, cycle, fatright = "←", "→", "⟳", "⮕"
    check, x = "✔", "✘"
    smile, frown, happy, worried, confused = "🙂", "☹", "😃", "😟", "😕"
    circle, square, triangle = "⚪", "◼", "▶"
    vline, hline, vdots = "|", "―", "⁞"
    bar, pipe, brokenbar, tech, zigzag = "―", "‖", "¦", "⌇", "⦚"
    # brackets
    langle, rangle = "⟨", "⟩"
    lshell, rshell = "⦗", "⦘"
    ldbracket, rdbracket = "⟦", "〛"
    ldshell, rdshell = "〘", "〙"
    ldparen, rdparen = "⸨", "⸩"
    ldangle, rdangle = "《", "》"
    # greek
    alpha, beta, gamma, delta, epsilon, eta, theta, zeta, kappa = (
        "α",
        "β",
        "γ",
        "δ",
        "ε",
        "η",
        "θ",
        "ζ",
        "κ",
    )
    Gamma, Delta, Pi, Sigma, Omega = "Γ", "Δ", "Π", "Σ", "Ω"
    lamba = "λ"  # spelled wrong
    nu, mu, xi, tau, pi, sigma, phi, psi, omega = (
        "ν",
        "μ",
        "ξ",
        "τ",
        "π",
        "σ",
        "φ",
        "ψ",
        "ω",
    )
    varphi = "φ"

    @classmethod
    def range(cls, start: Any, end: Any) -> str:
        return str(start) + cls.en + str(end)

    @classmethod
    def squoted(cls, s: Any) -> str:
        """Wrap a string in singsle quotes."""
        return Chars.lsq + str(s) + Chars.rsq

    @classmethod
    def dquoted(cls, s: Any) -> str:
        """Wrap a string in double quotes."""
        return Chars.ldq + str(s) + Chars.rdq

    @classmethod
    def angled(cls, s: Any) -> str:
        """Wrap a string in angled brackets."""
        return Chars.langle + str(s) + Chars.rangle

    @classmethod
    def dangled(cls, s: Any) -> str:
        """Wrap a string in double brackets."""
        return Chars.ldangle + str(s) + Chars.rdangle

    @classmethod
    def parened(cls, s: Any) -> str:
        """Wrap a string in parentheses."""
        return "(" + str(s) + ")"

    @classmethod
    def bracketed(cls, s: Any) -> str:
        """Wrap a string in square brackets."""
        return "[" + str(s) + "]"

    @classmethod
    def braced(cls, s: Any) -> str:
        """Wrap a string in curly braces."""
        return "{" + str(s) + "}"

    @classmethod
    def shelled(cls, s: Any) -> str:
        """Wrap a string in tortiose shell brackets (〔 〕)."""
        return "〔" + str(s) + "〕"

    @classmethod
    def dbracketed(cls, s: Any) -> str:
        """Wrap a string in double square brackets (⟦ ⟧)."""
        return Chars.ldbracket + str(s) + Chars.rdbracket


__all__ = ["Chars"]



from pocketutils.core import OptRow, SmartEnum, frozenlist
from pocketutils.core.chars import *
from pocketutils.core.exceptions import *
from pocketutils.core.input_output import *
from pocketutils.core.iterators import *
from pocketutils.tools.all_tools import Tools



from __future__ import annotations

import abc
import enum
import json
import logging
from datetime import date, datetime
from typing import Any, Callable, Generic, Iterable, Sequence, Type, TypeVar

import numpy as np

# noinspection PyProtectedMember
from pocketutils.core._internal import PathLike, PathLikeUtils
from pocketutils.core._internal import look as _look
from pocketutils.core.exceptions import ImmutableError

T = TypeVar("T", covariant=True)

logger = logging.getLogger("pocketutils")


class Sentinel:
    """
    A sentinel value tied to nothing more than a memory address.
    """

    @classmethod
    def new(cls) -> Sentinel:
        return Sentinel()

    def __init__(self):
        pass


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


V = TypeVar("V")


class LazyWrapped(Generic[V], metaclass=abc.ABCMeta):
    def __init__(self):
        self._v, self._exists = None, False

    def get(self) -> V:
        if not self._exists:
            self._v = self._generate()
            self._exists = True
        return self._v

    @property
    def raw_value(self):
        return self._v

    @property
    def is_defined(self):
        return self._exists

    @property
    def _name(self):
        raise NotImplementedError()

    def _generate(self):
        raise NotImplementedError()

    def __repr__(self):
        return self._name + "[" + (repr(self._v) if self.is_defined else "⌀") + "]"

    def __str__(self):
        return self._name + "[" + (str(self._v) if self.is_defined else "⌀") + "]"

    def __eq__(self, other):
        return (
            type(self) == type(other)
            and self.is_defined == other.is_defined
            and self.raw_value == other.raw_value
        )


class PlainLazyWrapped(LazyWrapped, metaclass=abc.ABCMeta):
    pass


class ClearableLazyWrapped(LazyWrapped, metaclass=abc.ABCMeta):
    def clear(self) -> None:
        self._exists = False


class LazyWrap:
    @classmethod
    def new_type(cls, dtype: str, generator: Callable[[], V]) -> Type[PlainLazyWrapped]:
        # noinspection PyTypeChecker
        return cls._new_type(dtype, generator, PlainLazyWrapped)

    @classmethod
    def new_clearable_type(
        cls, dtype: str, generator: Callable[[], V]
    ) -> Type[ClearableLazyWrapped]:
        # noinspection PyTypeChecker
        return cls._new_type(dtype, generator, ClearableLazyWrapped)

    @classmethod
    def _new_type(
        cls, dtype: str, generator: Callable[[], V], superclass: Type[LazyWrapped]
    ) -> Type[LazyWrapped]:
        """
        Creates a new mutable wrapped type.

        Example:
            >>> LazyRemoteTime = LazyWrap.new_type('RemoteTime', lambda: ...)
            >>> dt = LazyRemoteTime()  # nothing happens
            >>> dt.get()  # has a value

        Args:
            dtype: The name of the data type, such as 'datetime' if generator=datetime.now
            generator: This is called to (lazily) initialize an instance of the LazyWrapped

        Returns:
            A new class subclassing LazyWrapped
        """

        class X(superclass):
            @property
            def _name(self):
                return dtype

            def _generate(self):
                return generator()

        X.__name__ = superclass.__name__ + dtype
        return X


class SmartEnum(enum.Enum):
    """
    An enum with a classmethod ``of`` that parses a string of the member's name.
    """

    @classmethod
    def of(cls, v):
        """
        Returns the member of this enum class from a string with the member's name,
        case-insensitive and stripping whitespace.
        Will return ``v`` if ``v`` is already an instance of this class.
        """
        if isinstance(v, cls):
            return v
        elif isinstance(v, str):
            if v in cls:
                return cls[v.upper().strip()]
            else:
                # in case the names are lowercase
                # noinspection PyTypeChecker
                for e in cls:
                    if e.name.lower().strip() == v:
                        return e
                raise LookupError(f"{v} not found in {str(cls)}")
        else:
            raise TypeError(str(type(v)))


# noinspection PyPep8Naming
class frozenlist(Sequence):
    """
    An immutable sequence backed by a list.
    The sole advantage over a tuple is the list-like __str__ with square brackets, which may be less confusing to a user.
    """

    def __init__(self, items: Iterable[T]):
        self.__items = list(items)

    def __getitem__(self, item) -> T:
        if isinstance(item, int):
            return self.__items[item]
        else:
            return frozenlist(self.__items[item])

    def __setitem__(self, key, value):
        raise ImmutableError()

    def __len__(self) -> int:
        return len(self.__items)

    def __repr__(self):
        return repr(self.__items)

    def __eq__(self, other):
        return type(self) == type(other) and self.__items == other.__items

    def __str__(self):
        return repr(self.__items)


class OptRow:
    """
    Short for 'optional row'.
    A wrapper around a NamedTuple that returns None if the key doesn't exist.
    This is intended for Pandas itertuples().
    """

    def __init__(self, row):
        self._row = row

    def __getattr__(self, item: str) -> Any:
        try:
            return getattr(self._row, item)
        except AttributeError:
            return None

    def opt(self, item: str, look=None) -> Any:
        x = getattr(self, item)
        if x is None:
            return None
        return _look(x, look)

    def req(self, item: str, look=None) -> Any:
        x = getattr(self._row, item)
        if x is None:
            return None
        return _look(x, look)

    def __contains__(self, item):
        try:
            getattr(self._row, item)
            return True
        except AttributeError:
            return False

    def items(self):
        # noinspection PyProtectedMember
        return self._row._asdict()

    def keys(self):
        # noinspection PyProtectedMember
        return self._row._asdict().keys()

    def values(self):
        # noinspection PyProtectedMember
        return self._row._asdict().values()

    def __repr__(self):
        return self.__class__.__name__ + "@" + hex(id(self))

    def __str__(self):
        return self.__class__.__name__

    def __eq__(self, other):
        # noinspection PyProtectedMember
        return self._row == other._row


__all__ = [
    "Sentinel",
    "JsonEncoder",
    "SmartEnum",
    "frozenlist",
    "PathLike",
    "PathLikeUtils",
    "OptRow",
    "LazyWrap",
]



import typing
from abc import ABCMeta
from functools import total_ordering
from typing import Iterator, Sequence, Type

import regex

from pocketutils.core.exceptions import OutOfRangeError, StringPatternError


@total_ordering
class _WB(metaclass=ABCMeta):
    """
    A set of conversion utils between well labels, indices, and coordinates.
    This class is preferred over the functions above.
    Short for "Well Base".
    This superclass is hidden. Use only the subclasses WB0 and WB1.
    Limited to plates with no more than 26 rows.
    """

    def __init__(self, n_rows: int, n_columns: int):
        self.base = self.get_base()
        self.n_rows = n_rows
        self.n_columns = n_columns
        self.n_wells = n_rows * n_columns
        self.__well_name0 = self._ind_to_label
        self.__well_index0 = self._label_to_ind
        self.__well_rc0 = lambda i: (i // n_columns, i % n_columns)
        self.__rc_to_i0 = lambda r, c: self.n_columns * r + c

    @classmethod
    def get_base(cls) -> int:
        raise NotImplementedError()

    def __hash__(self):
        return hash((self.base, self.n_rows, self.n_columns))

    def __eq__(self, other):
        if not isinstance(other, _WB):
            raise TypeError(f"Cannot compare type {type(other)}")
        return (self.base, self.n_rows, self.n_columns) == (
            other.base,
            other.n_rows,
            other.n_columns,
        )

    def label_to_index(self, label: str) -> int:
        return self.__well_index0(label) + self.base

    def label_to_rc(self, label: str) -> typing.Tuple[int, int]:
        r, c = self.__well_rc0(self.__well_index0(label))
        return r + self.base, c + self.base

    def index_to_label(self, i: int) -> str:
        self.__check_index_range(i)
        return self.__well_name0(i - self.base)

    def index_to_rc(self, i: int) -> typing.Tuple[int, int]:
        self.__check_index_range(i)
        r, c = self.__well_rc0(i - self.base)
        return r + self.base, c + self.base

    def rc_to_label(self, row: int, column: int) -> str:
        self.__check_rc_range(row, column)
        i = self.__rc_to_i0(row - self.base, column - self.base)
        return self.__well_name0(i)

    def rc_to_index(self, row: int, column: int) -> int:
        self.__check_rc_range(row, column)
        return self.__rc_to_i0(row - self.base, column - self.base) + self.base

    def all_labels(self) -> Sequence[str]:
        return [self.index_to_label(i) for i in self.all_indices()]

    def all_rcs(self) -> Sequence[typing.Tuple[int, int]]:
        return [self.index_to_rc(i) for i in self.all_indices()]

    def all_indices(self) -> Sequence[int]:
        return list(range(self.base, self.n_rows * self.n_columns + self.base))

    def simple_range(self, a: str, b: str) -> Iterator[str]:
        ar, ac = self.label_to_rc(a)
        br, bc = self.label_to_rc(b)
        if ar == br:
            for c in range(ac, bc + 1):
                yield self.rc_to_label(ar, c)
        elif ac == bc:
            for r in range(ar, br + 1):
                yield self.rc_to_label(r, ac)
        else:
            raise OutOfRangeError(f"{a}-{b} is not a simple range")

    def block_range(self, a: str, b: str) -> Iterator[str]:
        ar, ac = self.label_to_rc(a)
        br, bc = self.label_to_rc(b)
        for r in range(ar, br + 1):
            for c in range(ac, bc + 1):
                yield self.rc_to_label(r, c)

    def traversal_range(self, a: str, b: str) -> Iterator[str]:
        ai = self.label_to_index(a)
        bi = self.label_to_index(b)
        for i in range(ai, bi + 1):
            yield self.index_to_label(i)

    def __check_rc_range(self, row: int, column: int):
        if (
            row < self.base
            or row > self.n_rows * self.n_columns
            or column < self.base
            or column > self.n_rows * self.n_columns + self.base - 1
        ):
            raise OutOfRangeError(f"{self.base}-based coordinates {(row, column)} out of range")

    def __check_index_range(self, i: int):
        if i < self.base or i > self.n_wells + self.base - self.base:
            raise OutOfRangeError(f"{self.base}-based index {i} out of range")

    def __lt__(self, other):
        if self.__class__ != other.__class__:
            raise TypeError(f"Wrong type {type(other)}")
        if self.n_wells < other.n_wells:
            return True
        if self.n_wells > other.n_wells:
            return False
        return (self.n_rows, self.n_columns) < (other.n_rows, other.n_columns)

    def __repr__(self):
        return "WB{}({}×{})".format(self.base, self.n_rows, self.n_columns)

    def __str__(self):
        return "WB{}({}×{})".format(self.base, self.n_rows, self.n_columns)

    def _ind_to_label(self, index: int) -> str:
        """Returns a 3-character well identifier like A52.
        Note that the default is landscape indexing (8 rows and 12 columns).

        Modified from https://stackoverflow.com/questions/19170420/converting-well-number-to-the-identifier-on-a-96-well-plate.
        """
        index = int(index)
        if (
            self.n_rows > 52
        ):  # a 1536-well plate has 32 rows and 48 columns, so this won't work with that
            raise OutOfRangeError("Well names are limited to plates with 26 rows!")
        if index >= self.n_rows * self.n_columns:
            raise OutOfRangeError(
                "Well index {} is out of range (max is {})".format(
                    index, self.n_rows * self.n_columns
                )
            )
        return [chr(i) for i in range(0x41, 0x41 + self.n_rows)][
            index // self.n_columns
        ] + "%02d" % (index % self.n_columns + 1,)

    def _label_to_ind(self, name: str) -> int:
        return (ord(name[0]) % 32) * self.n_columns - 1 - (self.n_columns - int(name[1:]))


class WbFactory:
    @classmethod
    def new(cls, base: int) -> Type[_WB]:
        new_class = type(f"WB{base}", (_WB,), {})
        new_class.get_base = lambda c: base
        # noinspection PyTypeChecker
        return new_class


class ParsingWB(_WB, metaclass=ABCMeta):
    """
    An abstract WB that can parse expressions.
    Usage:
        >>> class ParsingWB1(WB1, ParsingWB): pass
    """

    _pattern = regex.compile(
        r""" *([A-H][0-9]{1,2}) *(?:(-|–|\*|(?:\.\.\.)|…) *([A-H][0-9]{1,2}))? *""", flags=regex.V1
    )

    def parse(self, expression: str) -> Sequence[str]:
        """
        Returns the labels from the expression, inclusive.
        Examples:
            A01-A10   (sequence in a single row)
            A01-E01   (sequence in a single column)
            A01*C01   (a rectangular block)
            A01...C01 (a traversal of the wells in order)
        """
        trav = []
        for txt in expression.split(","):
            trav.extend(self._parse(txt))
        return trav

    def _parse(self, expression: str):
        match = ParsingWB._pattern.fullmatch(expression)
        if match is None:
            raise StringPatternError(
                f"{expression} is wrong",
                value=expression,
                pattern=ParsingWB._pattern,
            )
        a, x, b = match.group(1), match.group(2), match.group(3)
        if x is None:
            return self.simple_range(a, a)
        elif x in {"-", "–"}:
            return self.simple_range(a, b)
        elif x == "*":
            return self.block_range(a, b)
        elif x in {"...", "…"}:
            return self.traversal_range(a, b)
        else:
            raise ValueError(f"Failed on to parse {x}")


class WB1(_WB):
    """
    Map of a plate with conversions between labels, indices, and (row, column) pairs.
    ALL of these methods assume 1-based indices (1 is the first index and (1,1) is the first coordinate).
    """

    @classmethod
    def get_base(cls):
        return 1


class WB0(_WB):
    """
    Map of a plate with conversions between labels, indices, and (row, column) pairs.
    ALL of these methods assume 0-based indices (1 is the first index and (0,0) is the first coordinate).
    The labels still start at A1 (0,0).
    """

    @classmethod
    def get_base(cls):
        return 0


class ParsingWB0(WB0, ParsingWB):
    pass


class ParsingWB1(WB1, ParsingWB):
    pass


ParsingWB0.__doc__ = WB0.__doc__
ParsingWB1.__doc__ = WB1.__doc__

__all__ = ["WB1", "WB0", "ParsingWB0", "ParsingWB1", "WbFactory"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pocketutils',
 'pocketutils.biochem',
 'pocketutils.core',
 'pocketutils.misc',
 'pocketutils.plotting',
 'pocketutils.tools']

package_data = \
{'': ['*']}

install_requires = \
['orjson>=3.4,<4.0', 'regex>=2021', 'toml>=0.10,<1.0', 'tomlkit>=0.7,<1.0']

extras_require = \
{'all': ['loguru>=0.5,<1.0',
         'defusedxml>=0.7,<1.0',
         'dill>=0.3,<1.0',
         'jsonpickle>=2,<3',
         'joblib>=1,<2',
         'numpy>=1.19,<2.0',
         'pandas>=1.2,<2.0',
         'pint>=0.17,<1.0',
         'matplotlib>=3.3,<4.0',
         'goatools>=1,<2',
         'requests>=2,<3',
         'uniprot>=1.3,<2.0',
         'colorama>=0.4,<1.0',
         'psutil>=5,<6',
         'typer>=0.4,<1.0',
         'ipython>=7.0'],
 'biochem': ['numpy>=1.19,<2.0',
             'pandas>=1.2,<2.0',
             'pint>=0.17,<1.0',
             'goatools>=1,<2',
             'requests>=2,<3',
             'uniprot>=1.3,<2.0'],
 'misc': ['colorama>=0.4,<1.0', 'psutil>=5,<6', 'typer>=0.4,<1.0'],
 'notebooks': ['pandas>=1.2,<2.0', 'ipython>=7.0'],
 'plotting': ['numpy>=1.19,<2.0', 'pandas>=1.2,<2.0', 'matplotlib>=3.3,<4.0'],
 'tools': ['loguru>=0.5,<1.0',
           'defusedxml>=0.7,<1.0',
           'dill>=0.3,<1.0',
           'jsonpickle>=2,<3',
           'joblib>=1,<2',
           'numpy>=1.19,<2.0',
           'pandas>=1.2,<2.0',
           'pint>=0.17,<1.0']}

setup_kwargs = {
    'name': 'pocketutils',
    'version': '0.7.0',
    'description': 'Adorable little Python code for you to copy or import.',
    'long_description': '# pocketutils\n\n[![Version status](https://img.shields.io/pypi/status/pocketutils?label=status)](https://pypi.org/project/pocketutils)\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)\n[![Python version compatibility](https://img.shields.io/pypi/pyversions/pocketutils?label=Python)](https://pypi.org/project/pocketutils)\n[![Version on Docker Hub](https://img.shields.io/docker/v/dmyersturnbull/pocketutils?color=green&label=Docker%20Hub)](https://hub.docker.com/repository/docker/dmyersturnbull/pocketutils)\n[![Version on Github](https://img.shields.io/github/v/release/dmyersturnbull/pocketutils?include_prereleases&label=GitHub)](https://github.com/dmyersturnbull/pocketutils/releases)\n[![Version on PyPi](https://img.shields.io/pypi/v/pocketutils?label=PyPi)](https://pypi.org/project/pocketutils)  \n[![Build (Actions)](https://img.shields.io/github/workflow/status/dmyersturnbull/pocketutils/Build%20&%20test?label=Tests)](https://github.com/dmyersturnbull/pocketutils/actions)\n[![Documentation status](https://readthedocs.org/projects/pocketutils/badge)](https://pocketutils.readthedocs.io/en/stable/)\n[![Coverage (coveralls)](https://coveralls.io/repos/github/dmyersturnbull/pocketutils/badge.svg?branch=main&service=github)](https://coveralls.io/github/dmyersturnbull/pocketutils?branch=main)\n[![Maintainability (Code Climate)](https://api.codeclimate.com/v1/badges/eea2b741dbbbb74ad18a/maintainability)](https://codeclimate.com/github/dmyersturnbull/pocketutils/maintainability)\n[![Scrutinizer Code Quality](https://scrutinizer-ci.com/g/dmyersturnbull/pocketutils/badges/quality-score.png?b=main)](https://scrutinizer-ci.com/g/dmyersturnbull/pocketutils/?branch=main)\n\nAdorable little Python functions for you to copy or import.\n\n`pip install pocketutils` or\n`pip install pocketutils[all]`\n[Apache](https://spdx.org/licenses/Apache-2.0.html)-licensed.\n\n### Basic usage – `Tools`\n\n```python\nfrom pocketutils.full import Tools\n\nTools.zip_strict([1, 2, 3], [5, 6])  # error <-- lengths must match\nTools.strip_brackets("( (xy)")  # "(xy" <-- strips paired only\nTools.sanitize_path("x\\ty")  # "xy"  <-- very robust cross-platform sanitization\nTools.delete_surefire("my_file")  # <-- Attempts to fix permissions if needed\nTools.git_description("my_repo").tag  # <-- get git repo info\nTools.pretty_function(lambda s: None)  # "<λ(1)> <-- decent name for any object\nTools.roman_to_arabic("XIV")  # 14  <-- inverse function too\nTools.delta_time_to_str(delta_sec=60 * 2 + 5)  # "02:05"  <-- handles days too\nTools.round_to_sigfigs(135.3, 2)  # 140  <-- rounding to sigfigs-proper\nTools.pretty_float(-float("-inf"))  # "−∞"  <-- proper unicode, no trailing 0s\nTools.stream_cmd_call(["cat", "big-file"], callback=fn)  # <-- buffer never fills\nTools.strip_off("hippopotamus", "hippo")  # "potamus"  <-- what .strip() should do\nTools.strip_quotes("\'hello\'")  # "hello"\nTools.truncate10("looong string")  # "looong st…"\nTools.parse_bool("true")  # True\nTools.parse_bool_flex("yes")  # True\nTools.look(item, "purchase.buyer.first_name")  # None if purchase or buyer is None\nTools.friendly_size(n_bytes=2 * 14)  # "16.38 kb"\nTools.is_probable_null("NaN")  # True\nTools.is_true_iterable("kitten")  # False\nTools.or_null(some_function)  # None if it fails\nTools.or_raise(None)  # raises an error (of your choice)\nTools.trash(unwanted_file)  # move to os-specific trash\nTools.pretty_dict({"contents": {"greeting": "hi"}})  # indented\nTools.save_diagnostics(Tools.get_env_info())  # record diagnostic info\nTools.is_lambda(lambda: None)  # True\nTools.longest(["a", "a+b"])  # "a+b"  # anything with len\nTools.only([1, 2])  # error -- multiple items\nTools.first(iter([]))  # None <-- better than try: next(iter(x)) except:...\n# lots of others\n```\n\n### More things\n\n- `FancyLoguru` (really useful)\n- `NestedDotDict` (esp. for toml and json)\n- `QueryUtils` (handles rate-limiting, etc.)\n- `FigTools` (for matplotlib)\n- `J` (tools to interact with Jupyter)\n- `WB1` (microwell plate nomenclature)\n- `Chars` (e.g. `Chars.shelled(s)` or `Chars.snowflake`)\n- `exceptions` (general-purpose exceptions that can store relevant info)\n\n_Even more, albeit more obscure:_\n\n- `TissueExpression`, `UniprotGo`, `AtcTree`, `PlateRois`\n- `WebResource`, `magic_template`\n- `color_schemes`, `FigSaver`, `RefDims`\n- `LoopTools`\n- `MemCache`\n\n[See the docs 📚](https://pocketutils.readthedocs.io/en/stable/), or just\n[browse the code](https://github.com/dmyersturnbull/pocketutils/tree/main/pocketutils).\n[New issues](https://github.com/dmyersturnbull/pocketutils/issues) and pull requests are welcome.\nPlease refer to the [contributing guide](https://github.com/dmyersturnbull/pocketutils/blob/main/CONTRIBUTING.md)\nand [security policy](https://github.com/dmyersturnbull/pocketutils/blob/main/SECURITY.md).  \nGenerated with tyrannosaurus: `tyrannosaurus new tyrannosaurus`\n',
    'author': 'Douglas Myers-Turnbull',
    'author_email': None,
    'maintainer': 'Douglas Myers-Turnbull',
    'maintainer_email': None,
    'url': 'https://github.com/dmyersturnbull/pocketutils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
import operator
import os
import sys
from pathlib import PurePath
from typing import Any, Callable, Iterable, Optional, TypeVar, Union

T = TypeVar("T", covariant=True)
Y = TypeVar("Y")
Z = TypeVar("Z")
logger = logging.getLogger("pocketutils")

PathLike = Union[str, PurePath, os.PathLike]


class PathLikeUtils:
    @classmethod
    def isinstance(cls, value: Any):
        return (
            isinstance(value, str) or isinstance(value, os.PathLike) or isinstance(value, PurePath)
        )


class Pretty:
    @classmethod
    def condensed(cls, item, depth=1):
        if isinstance(item, dict):
            return (
                "{\n"
                + "\n".join(
                    [
                        "\t" * (depth + 1) + k + " = " + cls.condensed(v, depth + 1)
                        for k, v in item.items()
                    ]
                )
                + "\n"
                + "\t" * depth
                + "}"
            )
        else:
            return str(item)

    @classmethod
    def expanded(cls, item, depth=1):
        if isinstance(item, dict):
            return (
                "{\n"
                + "\n".join(
                    [
                        "\t" * (depth + 1) + k + " = " + cls.expanded(v, depth + 1)
                        for k, v in item.items()
                    ]
                )
                + "\n"
                + "\t" * depth
                + "}"
            )
        elif isinstance(item, (list, set)):
            return (
                "[\n"
                + "\n".join(["\t" * (depth + 1) + cls.expanded(v, depth + 1) for v in item])
                + "\n"
                + "\t" * depth
                + "]"
            )
        else:
            return str(item)


def nicesize(nbytes: int, space: str = "") -> str:
    """
    Uses IEC 1998 units, such as KiB (1024).
        nbytes: Number of bytes
        space: Separator between digits and units

        Returns:
            Formatted string
    """
    data = {
        "PiB": 1024 ** 5,
        "TiB": 1024 ** 4,
        "GiB": 1024 ** 3,
        "MiB": 1024 ** 2,
        "KiB": 1024 ** 1,
    }
    for suffix, scale in data.items():
        if nbytes >= scale:
            break
    else:
        scale, suffix = 1, "B"
    return str(nbytes // scale) + space + suffix


def look(obj: Y, attrs: Union[str, Iterable[str], Callable[[Y], Z]]) -> Optional[Z]:
    if attrs is None:
        return obj
    if not isinstance(attrs, str) and hasattr(attrs, "__len__") and len(attrs) == 0:
        return obj
    if isinstance(attrs, str):
        attrs = operator.attrgetter(attrs)
    elif isinstance(attrs, Iterable) and all((isinstance(a, str) for a in attrs)):
        attrs = operator.attrgetter(".".join(attrs))
    elif not callable(attrs):
        raise TypeError(
            f"Type {type(attrs)} unrecognized for key/attrib. Must be a function, string, or sequence of strings"
        )
    try:
        return attrs(obj)
    except AttributeError:
        return None


def null_context():
    yield


__all__ = ["nicesize", "look", "logger", "Pretty", "PathLike", "PathLikeUtils"]



from __future__ import annotations

import colorsys
from typing import Sequence

from matplotlib import colors as mcolors
from matplotlib.colors import Colormap, LinearSegmentedColormap


class FancyColorSchemes:
    """
    Color schemes that can be used.
    Each is a list of hex color codes.
    The "tol" palettes (ex ``qualitative_tol_vibrant_7``) are safe for red-green color blindness.
    For all palettes other than grayscale, colors of black or gray were removed.
    """

    @classmethod
    def darken_color(cls, color: str, shading: float) -> str:
        """
        Darkens a color by a certain fraction.

        Args:
            color: A matplotlib-recognized colors
            shading: A float from 0 to 1, with 1 being complete shading
        """
        color = mcolors.colorConverter.to_rgb(color)
        h, l, s = colorsys.rgb_to_hls(*mcolors.to_rgb(color))
        rgb = colorsys.hls_to_rgb(h, shading * 1, s)
        return mcolors.to_hex(rgb)

    @classmethod
    def darken_palette(cls, palette: Sequence[str], shading: float) -> Sequence[str]:
        """
        Darkens a list of colors by a certain fraction.

        Args:
            palette: A sequence of matplotlib-recognized colors
            shading: A float from 0 to 1, with 1 being complete shading
        """
        return [FancyColorSchemes.darken_color(c, shading) for c in palette]

    @classmethod
    def grayscales(cls) -> Sequence[str]:
        """
        A range of grayscale-ISH colors from black to white.
        Some have slightly more of red, green, or blue.
        """
        return [
            "#000000",
            "#555555",
            "#999999",
            "#cccccc",
            "#300000",
            "#000030",
            "#003333",
            "#333300",
            "#775555",
            "#555577",
            "#557777",
            "#777755",
        ]

    @classmethod
    def qualitiative_26(cls) -> Sequence[str]:
        """Special color scheme of 26."""
        # the first 11 of these were *[c for c in InternalKvrcUtils.darken_palette(qualitative.Set3_12.hex_colors, 0.38) if c != '#787878']
        # some were slightly modified to work better with the rest
        return [
            "#367c7d",
            "#daca00",
            "#44337e",
            "#bc1805",
            "#32699a",
            "#cc6016",
            "#719f23",
            "#b70b63",
            "#816161",
            "#7e007e",
            "#ff9000",
            "#c2aa00",
            "#11e022",
            "#0000ff",
            "#00bb00",
            "#fb9a99",
            "#ff0000",
            "#cc8700",
            "#006400",
            "#554411",
            "#191970",
            "#00aa66",
            "#dd6666",
            "#2288ee",
            "#009900",
            "#c71588",
        ]

    @classmethod
    def qualitative_tol_dark_mod_6(cls) -> Sequence[str]:
        """
        "Dark" color scheme from https://personal.sron.nl/~pault/#sec:qualitative
        No black and reordered red and yellow.
        """
        return ["#222255", "#225555", "#225522", "#663333", "#666633", "#551144"]

    @classmethod
    def qualitative_tol_high_contrast_6(cls) -> Sequence[str]:
        """
        "High-contrast" color scheme from https://personal.sron.nl/~pault/#sec:qualitative
        Reordered. No gray.
        """
        return ["#0077bb", "#ee3377", "#33bbee", "#cc3311", "#009988", "#ee7733"]

    @classmethod
    def qualitative_tol_vibrant_6(cls) -> Sequence[str]:
        """
        "Vibrant" color scheme from https://personal.sron.nl/~pault/#sec:qualitative
        Gray was removed.
        """
        return ["#0077BB", "#CC3311", "#009988", "#EE7733", "#33BBEE", "#EE3377"]

    @classmethod
    def qualitative_tol_muted_9(cls) -> Sequence[str]:
        """
        "Muted" color scheme from https://personal.sron.nl/~pault/#sec:qualitative
        Gray was removed.

        Returns:

        """
        return [
            "#332288",
            "#88CCEE",
            "#44AA99",
            "#117733",
            "#999933",
            "#DDCC77",
            "#CC6677",
            "#882255",
            "#AA4499",
        ]


class FancyCmaps:
    """
    Very useful colormaps. Most importantly:
        - white_red
        - white_blue
        - blue_white
        - white_black

    The built-in matplotlib ones don't actually go between pure color values!!
    For ex, 'Greys' doesn't go from pure white to pure black!
    So colormaps to consider avoiding include Greys, Blues, Greens, (etc), bwr, and seismic.
    Matplotlib still has good built-in colormaps, including viridis and plasma.
    """

    @classmethod
    def white_red(cls, *, bad: str = "333333") -> Colormap:
        """A colormap from pure white to pure red."""
        cmap = LinearSegmentedColormap.from_list("white_red", ["#ffffff", "#ff0000"])
        cmap.set_bad(color=bad)
        return cmap

    @classmethod
    def white_blue(cls, *, bad: str = "333333") -> Colormap:
        """A colormap from pure white to pure blue."""
        cmap = LinearSegmentedColormap.from_list("white_blue", ["#ffffff", "#0000ff"])
        cmap.set_bad(color=bad)
        return cmap

    @classmethod
    def blue_white(cls, *, bad: str = "333333") -> Colormap:
        """A colormap from pure white to pure blue."""
        cmap = LinearSegmentedColormap.from_list("blue_white", ["#0000ff", "#ffffff"])
        cmap.set_bad(color=bad)
        return cmap

    @classmethod
    def white_black(cls, *, bad: str = "333333") -> Colormap:
        """A colormap from pure white to pure black."""
        cmap = LinearSegmentedColormap.from_list("white_black", ["#ffffff", "#000000"])
        cmap.set_bad(color=bad)
        return cmap

    @classmethod
    def blue_white_red(cls, *, bad: str = "333333") -> Colormap:
        """A colormap from pure blue to pure red."""
        cmap = LinearSegmentedColormap.from_list(
            "blue_white_red", ["#0000ff", "#ffffff", "#ff0000"]
        )
        cmap.set_bad(color=bad)
        return cmap


__all__ = ["FancyColorSchemes", "FancyCmaps"]



from __future__ import annotations

from typing import Callable, Optional

import pandas as pd

from pocketutils.core import PathLike
from pocketutils.core.exceptions import LookupFailedError
from pocketutils.core.web_resource import WebResource


class TissueTable(pd.DataFrame):
    """
    Contains a Pandas DataFrame of tissue- and cell type-level expression for genes from the Human Protein Atlas.

    Example:
        tt = TissueTable()
        tt.tissue('MKNK2') # returns a DataFrame with mean expression of MKNK2 per tissue type. MKN2 is the HGNC symbol.
    """

    URL = "https://www.proteinatlas.org/download/normal_tissue.tsv.zip"
    MEMBER_NAME = "normal_tissue.tsv"
    DEFAULT_PATH = "normal_tissue.tsv.gz"

    @classmethod
    def load(
        cls,
        path: Optional[PathLike] = None,
        filter_fn: Callable[[pd.DataFrame], pd.DataFrame] = pd.DataFrame.dropna,
    ) -> TissueTable:
        """
        Gets a DataFrame of Human Protein Atlas tissue expression data,
        indexed by Gene name and with the 'Gene' and 'Reliability' columns dropped.
        The expression level ('Level') is replaced using this map: {'Not detected': 0, 'Low': 1, 'Medium': 2, 'High': 3}.
        Downloads the file from http://www.proteinatlas.org/download/normal_tissue.tsv.zip
        and reloads from normal_tissue.tsv.gz thereafter.
        """
        if path is None:
            path = TissueTable.DEFAULT_PATH
        resource = WebResource(TissueTable.URL, TissueTable.MEMBER_NAME, path)
        resource.download()
        tissue = pd.read_csv(path, sep="\t")
        tissue = tissue.drop("Gene", axis=1).drop("Reliability", axis=1)
        tissue = filter_fn(tissue)
        tissue["Level"] = (
            tissue["Level"]
            .map({"Not detected": 0, "Low": 1, "Medium": 2, "High": 3}.get)
            .astype(float)
        )
        return TissueTable(tissue.set_index("Gene name"))

    def level(self, gene_name: str, group_by: str) -> TissueTable:
        """Returns a DataFrame of the mean expression levels by tissue or cell type."""
        if gene_name not in self.index.get_level_values("Gene name"):
            raise LookupFailedError(f"Gene with HGNC symbol {gene_name} not found.")
        gene = self[self.index.get_level_values("Gene name") == gene_name]
        if gene is None:
            raise AssertionError(f"Gene is None for gene_name {gene_name}, groupby {group_by}")
        return TissueTable(gene.groupby(group_by).mean().sort_values("Level", ascending=False))

    def tissue(self, name: str) -> TissueTable:
        return self.level(name, group_by="Tissue")

    def cell_type(self, name: str) -> TissueTable:
        return self.level(name, group_by="Cell type")


__all__ = ["TissueTable"]



import sys
import warnings
from typing import Callable, Mapping, Optional, Sequence

import regex

from pocketutils.core.exceptions import *
from pocketutils.tools.base_tools import BaseTools

logger = logging.getLogger("pocketutils")


class PathTools(BaseTools):
    @classmethod
    def updir(cls, n: int, *parts) -> Path:
        """
        Get an absolute path ``n`` parents from ``os.getcwd()``.
        Does not sanitize.

        Ex: In dir '/home/john/dir_a/dir_b':
            updir(2, 'dir1', 'dir2')  # returns Path('/home/john/dir1/dir2')
        """
        base = Path(os.getcwd())
        for _ in range(n):
            base = base.parent
        for part in parts:
            base = base / part
        return base.resolve()

    @classmethod
    def guess_trash(cls) -> Path:
        """
        Chooses a reasonable path for trash based on the OS.
        This is not reliable. For a more sophisticated solution,
        see https://github.com/hsoft/send2trash
        However, even that can fail.
        """
        plat = sys.platform.lower()
        if "darwin" in plat:
            return Path.home() / ".Trash"
        elif "win" in plat:
            return Path(Path.home().root) / "$Recycle.Bin"
        else:
            return Path.home() / ".trash"

    @classmethod
    def prep_dir(cls, path: PathLike, exist_ok: bool = True) -> bool:
        """
        Prepares a directory by making it if it doesn't exist.
        If exist_ok is False, calls logger.warning it already exists
        """
        path = Path(path)
        exists = path.exists()
        # On some platforms we get generic exceptions like permissions errors,
        # so these are better
        if exists and not path.is_dir():
            raise DirDoesNotExistError(f"Path {path} exists but is not a file")
        if exists and not exist_ok:
            logger.warning(f"Directory {path} already exists")
        if not exists:
            # NOTE! exist_ok in mkdir throws an error on Windows
            path.mkdir(parents=True)
        return exists

    @classmethod
    def prep_file(cls, path: PathLike, exist_ok: bool = True) -> None:
        """
        Prepares a file path by making its parent directory.
        Same as ``pathlib.Path.mkdir`` but makes sure ``path`` is a file if it exists.
        """
        # On some platforms we get generic exceptions like permissions errors, so these are better
        path = Path(path)
        # check for errors first; don't make the dirs and then fail
        if path.exists() and not path.is_file() and not path.is_symlink():
            raise FileDoesNotExistError(f"Path {path} exists but is not a file")
        Path(path.parent).mkdir(parents=True, exist_ok=exist_ok)

    @classmethod
    def sanitize_path(
        cls,
        path: PathLike,
        is_file: Optional[bool] = None,
        *,
        show_warnings: Union[bool, Callable[[str], Any]] = True,
    ) -> Path:
        r"""
        Sanitizes a path for major OSes and filesystems.
        Also see sanitize_path_nodes and sanitize_path_node.
        Platform-dependent.
        A corner case is drive letters in Linux:
        "C:\\Users\\john" is converted to '/C:/users/john' if os.name=='posix'
        """
        # the idea is to sanitize for both Windows and Posix, regardless of the platform in use
        # the sanitization should be as uniform as possible for both platforms
        # this works for at least Windows+NTFS
        # tilde substitution for long filenames in Windows -- is unsupported
        w = {True: logger.warning, False: lambda _: None}.get(show_warnings, show_warnings)
        if path.startswith("\\\\?"):
            w(f"Long UNC Windows paths (\\\\? prefix) are not supported (path '{path}')", path=path)
        bits = str(path).strip().replace("\\", "/").split("/")
        new_path = cls.sanitize_path_nodes(bits, is_file=is_file)
        if new_path != path:
            w(f"Sanitized filename {path} → {new_path}")
        return Path(new_path)

    @classmethod
    def sanitize_path_nodes(cls, bits: Sequence[PathLike], is_file: Optional[bool] = None) -> Path:
        fixed_bits = [
            bit + os.sep
            if i == 0 and bit.strip() in ["", ".", ".."]
            else cls.sanitize_path_node(
                bit,
                is_file=(False if i < len(bits) - 1 else is_file),
                is_root_or_drive=(None if i == 0 else False),
            )
            for i, bit in enumerate(bits)
            if bit.strip() not in ["", "."]
            or i == 0  # ignore // (empty) just like Path does (but fail on sanitize_path_node(' '))
        ]
        fixed_bits = [bit for i, bit in enumerate(fixed_bits) if i == 0 or bit not in ["", "."]]
        # unfortunately POSIX turns Path('C:\', '5') into C:\/5
        # this isn't an ideal way to fix it, but it works
        pat = regex.compile(r"^([A-Z]:)(?:\\)?$", flags=regex.V1)
        if os.name == "posix" and len(fixed_bits) > 0 and pat.fullmatch(fixed_bits[0]):
            fixed_bits[0] = fixed_bits[0].rstrip("\\")
            fixed_bits.insert(0, "/")
        return Path(*fixed_bits)

    @classmethod
    def sanitize_path_node(
        cls,
        bit: PathLike,
        is_file: Optional[bool] = None,
        is_root_or_drive: Optional[bool] = None,
        include_fat: bool = False,
    ) -> str:
        r"""
        Sanitizes a path node such that it will be fine for major OSes and filesystems.
        For example:
        - 'plums;and/or;apples' becomes 'plums_and_or_apples' (escaped ; and /)
        - 'null.txt' becomes '_null_.txt' ('null' is forbidden in Windows)
        - 'abc  ' becomes 'abc' (no trailing spaces)
        The behavior is platform-independent -- os, sys, and pathlib are not used.
        For ex, calling sanitize_path_node(r'C:\') returns r'C:\' on both Windows and Linux
        If you want to sanitize a whole path, see sanitize_path instead.

        Args:
            bit: The node
            is_file: False for directories, True otherwise, None if unknown
            is_root_or_drive: True if known to be the root ('/') or a drive ('C:\'), None if unknown
            include_fat: Also make compatible with FAT filesystems

        Returns:
            A string
        """
        # since is_file and is_root_or_drive are both Optional[bool], let's be explicit and use 'is' for clarity
        if is_file is True and is_root_or_drive is True:
            raise ContradictoryRequestError("is_file and is_root_or_drive are both true")
        if is_file is True and is_root_or_drive is None:
            is_root_or_drive = False
        if is_root_or_drive is True and is_file is None:
            is_file = False
        source_bit = copy(str(bit))
        bit = str(bit).strip()
        # first, catch root or drive as long as is_root_or_drive is not false
        # if is_root_or_drive is True (which is a weird call), then fail if it's not
        # otherwise, it's not a root or drive letter, so keep going
        if is_root_or_drive is not False:
            # \ is allowed in Windows
            if bit in ["/", "\\"]:
                return bit
            m = regex.compile(r"^([A-Z]:)(?:\\)?$", flags=regex.V1).fullmatch(bit)
            # this is interesting
            # for bit=='C:' and is_root_or_drive=None,
            # it could be either a drive letter
            # or a file path that should be corrected to 'C_'
            # I guess here we're going with a drive letter
            if m is not None:
                # we need C:\ and not C: because:
                # Path('C:\\', '5').is_absolute() is True
                # but Path('C:', '5').is_absolute() is False
                # unfortunately, doing Path('C:\\', '5') on Linux gives 'C:\\/5'
                # I can't handle that here, but sanitize_path() will account for it
                return m.group(1) + "\\"
            if is_root_or_drive is True:
                raise IllegalPathError(f"Node '{bit}' is not the root or a drive letter")
        # note that we can't call WindowsPath.is_reserved because it can't be instantiated on non-Linux
        # also, these appear to be different from the ones defined there
        bad_chars = {
            "<",
            ">",
            ":",
            '"',
            "|",
            "?",
            "*",
            "\\",
            "/",
            *{chr(c) for c in range(128, 128 + 33)},
            *{chr(c) for c in range(0, 32)},
            "\t",
        }
        # don't handle Long UNC paths
        # also cannot be blank or whitespace
        # the $ suffixed ones are for FAT
        # no CLOCK$, even with an ext
        # also no SCREEN$
        bad_strs = {
            "CON",
            "PRN",
            "AUX",
            "NUL",
            "COM1",
            "COM2",
            "COM3",
            "COM4",
            "COM5",
            "COM6",
            "COM7",
            "COM8",
            "COM9",
            "LPT1",
            "LPT2",
            "LPT3",
            "LPT4",
            "LPT5",
            "LPT6",
            "LPT7",
            "LPT8",
            "LPT9",
        }
        if include_fat:
            bad_strs += {"$IDLE$", "CONFIG$", "KEYBD$", "SCREEN$", "CLOCK$", "LST"}
        # just dots is invalid
        if set(bit.replace(" ", "")) == "." and bit not in ["..", "."]:
            raise IllegalPathError(f"Node '{source_bit}' is invalid")
        for q in bad_chars:
            bit = bit.replace(q, "_")
        if bit.upper() in bad_strs:
            # arbitrary decision
            bit = "_" + bit + "_"
        else:
            stub, ext = os.path.splitext(bit)
            if stub.upper() in bad_strs:
                bit = "_" + stub + "_" + ext
        if bit.strip() == "":
            raise IllegalPathError(f"Node '{source_bit}' is empty or contains only whitespace")
        # do this after
        if len(bit) > 254:
            raise IllegalPathError(f"Node '{source_bit}' has more than 254 characters")
        bit = bit.strip()
        if is_file is not True and (bit == "." or bit == ".."):
            return bit
        # never allow '.' (or ' ') to end a filename
        bit = bit.rstrip(".")
        return bit


__all__ = ["PathTools"]



import logging
import sys
import time
import warnings
from typing import Callable, Optional, Union

from pocketutils.tools.base_tools import BaseTools
from pocketutils.tools.filesys_tools import FilesysTools

logger = logging.getLogger("pocketutils")


class ConsoleTools(BaseTools):

    CURSOR_UP_ONE = "\x1b[1A"
    ERASE_LINE = "\x1b[2K"

    @classmethod
    def prompt_yes_no(cls, msg: str, writer: Callable[[str], None] = sys.stdout.write) -> bool:
        """
        Asks for "yes" or "no" via ``input``.
        Consider using ``typer.prompt`` instead.
        """
        while True:
            writer(msg + " ")
            command = input("")
            if command.lower() == "yes":
                return True
            elif command.lower() == "no":
                return False
            else:
                writer("Enter 'yes' or 'no'.\n")

    @classmethod
    def slow_delete(
        cls,
        path: str,
        wait: int = 5,
        *,
        delete_fn: Callable[[str], None] = FilesysTools.delete_surefire,
        writer: Callable[[str], None] = sys.stdout.write,
    ):
        logger.debug(f"Deleting directory tree {path} ...")
        writer(f"Waiting for {wait}s before deleting {path}: ")
        for i in range(0, wait):
            time.sleep(1)
            writer(str(wait - i) + " ")
        time.sleep(1)
        writer("...")
        delete_fn(path)
        writer(" deleted.\n")
        logger.debug(f"Deleted directory tree {path}")

    @classmethod
    def confirm(
        cls,
        msg: Union[None, str, Callable[[], None]] = None,
        *,
        input_fn: Callable[[str], str] = input,
        writer: Callable[[str], None] = sys.stdout.write,
    ) -> bool:
        """
        Asks for a confirmation from the user using the builtin ``input``.
        Consider using ``typer.prompt`` instead.
            msg: If None defaults to 'Confirm? [yes/no]'
            input_fn: Function to get the user input (its argument is always '')
            writer: Print using this function (should not print a newline by default)

        Returns:
            True if the user answered 'yes'; False otherwise
        """
        if msg is None:
            msg = "Confirm? [yes/no]"
        if isinstance(msg, str):

            def msg():
                writer(msg + " ")

        while True:
            msg()
            command = input_fn("").lower()
            if command in ["yes", "y"]:
                return True
            elif command in ["no", "n"]:
                return False

    @classmethod
    def clear_line(cls, n: int = 1, writer: Callable[[str], None] = sys.stdout.write) -> None:
        """
        Writes control characters to stdout to delete the previous line and move the cursor up.
        This only works in a shell.
            n: The number of lines to erase
            writer Write here
        """
        for _ in range(n):
            writer(ConsoleTools.CURSOR_UP_ONE)
            writer(ConsoleTools.ERASE_LINE)


__all__ = ["ConsoleTools"]



import typing
from dataclasses import dataclass
from enum import Enum
from typing import Dict

from pocketutils.core.exceptions import ErrorUtils, OutOfRangeError


class Edge(Enum):
    left = 1
    right = 2
    top = 3
    bottom = 4


class Axis(Enum):
    horizontal = 1
    vertical = 2


@ErrorUtils.args(edge=Edge, axis=Axis)
class RoiError(OutOfRangeError):
    """The ROI is invalid."""


@dataclass(frozen=True, order=True)
class WellRoi:
    row: int
    column: int
    x0: int
    y0: int
    x1: int
    y1: int

    def __post_init__(self):
        if self.row < 0:
            raise OutOfRangeError(f"Row ({self.row}) < 0", value=self.row)
        if self.column < 0:
            raise OutOfRangeError(f"Column ({self.column}) < 0", value=self.column)
        if self.x0 < 0:
            raise RoiError(f"x0 ({self.x0}) < 0", edge=Edge.left)
        if self.y0 < 0:
            raise RoiError(f"y0 ({self.y0}) < 0", edge=Edge.top)
        if self.x0 >= self.x1:
            raise RoiError(f"x0 ({self.x0}) >= x1 ({self.x1})", axis=Axis.horizontal)
        if self.y0 >= self.y1:
            raise RoiError(f"y0 ({self.y0}) >= y1 ({self.y1})", axis=Axis.vertical)

    def __repr__(self) -> str:
        return f"{self.row},{self.column}=({self.x0},{self.y0})→({self.x1},{self.y1})"

    def __str__(self):
        return repr(self)


@dataclass(frozen=True, repr=True, order=True)
class PlateRois:
    n_rows: int
    n_columns: int
    image_roi: WellRoi
    top_left_roi: WellRoi
    padx: float
    pady: float

    @property
    def well_rois(self):
        return self._get_roi_coordinates(self.top_left_roi, self.padx, self.pady)

    def __iter__(self):
        return iter(self.well_rois.keys())

    def __len__(self):
        return len(self.well_rois)

    def __getitem__(self, item):
        try:
            return self.well_rois[item[0], item[1]]
        except (IndexError, AttributeError):
            raise TypeError("Must look up well ROIs by (row, column) tuple indices.")

    def _get_roi_coordinates(
        self, top_left_roi: WellRoi, padx: float, pady: float
    ) -> Dict[typing.Tuple[int, int], WellRoi]:
        tl = top_left_roi
        width = top_left_roi.x1 - top_left_roi.x0
        height = top_left_roi.y1 - top_left_roi.y0
        wells_x_edge = tl.x0 + self.n_columns * width + (self.n_columns - 1) * padx
        wells_y_edge = tl.y0 + self.n_rows * height + (self.n_rows - 1) * pady
        # make sure the wells don't extend outside the image bounds
        if tl.x0 < self.image_roi.x0:
            raise RoiError(f"{tl.x0} < {self.image_roi.x0}", edge=Edge.left)
        if wells_x_edge > self.image_roi.x1:
            raise RoiError(f"{wells_x_edge} < {self.image_roi.x1}", edge=Edge.right)
        if tl.y0 < self.image_roi.y0:
            raise RoiError(f"{tl.y0} < {self.image_roi.y0}", edge=Edge.top)
        if wells_y_edge > self.image_roi.y1:
            raise RoiError(f"{wells_y_edge} > {self.image_roi.y1}", edge=Edge.bottom)
        # now build
        rois = {}
        x = tl.x0
        y = tl.y0
        for row in range(0, self.n_rows):
            for column in range(0, self.n_columns):
                rois[(row, column)] = WellRoi(row, column, x, y, x + width, y + height)
                x += width + padx
            y += height + pady
            x = tl.x0
        return rois


__all__ = ["WellRoi", "PlateRois"]



import contextlib
import logging
import os
import subprocess  # nosec
import textwrap
import warnings
from copy import copy
from enum import Enum
from pathlib import PurePath
from queue import Queue
from threading import Thread
from typing import Callable, Generator, Optional, Sequence

from pocketutils.core.input_output import DevNull
from pocketutils.tools.base_tools import BaseTools

logger = logging.getLogger("pocketutils")


@contextlib.contextmanager
def null_context(cls):
    yield


class CallTools(BaseTools):
    @classmethod
    @contextlib.contextmanager
    def silenced(
        cls, no_stdout: bool = True, no_stderr: bool = True
    ) -> Generator[None, None, None]:
        """
        Context manager that suppresses stdout and stderr.
        """
        # noinspection PyTypeChecker
        with contextlib.redirect_stdout(DevNull()) if no_stdout else cls.null_context():
            # noinspection PyTypeChecker
            with contextlib.redirect_stderr(DevNull()) if no_stderr else cls.null_context():
                yield

    @classmethod
    def call_cmd(cls, *cmd: str, **kwargs) -> subprocess.CompletedProcess:
        """
        Calls subprocess.run with capture_output=True.
        Logs a debug statement with the command beforehand.
            cmd: A sequence to call
            kwargs: Passed to subprocess.run
        """
        warnings.warn("call_cmd will be removed; use subprocess.check_output instead")
        logger.debug("Calling '{}'".format(" ".join(cmd)))
        return subprocess.run(*[str(c) for c in cmd], capture_output=True, check=True, **kwargs)

    @classmethod
    def call_cmd_utf(
        cls, *cmd: str, log: logging.Logger = logger, **kwargs
    ) -> subprocess.CompletedProcess:
        """
        Like ``call_cmd`` for utf-8 only.
        Set ``text=True`` and ``encoding=utf8``,
        and strips stdout and stderr of start/end whitespace before returning.
        Can also log formatted stdout and stderr on failure.
        Otherwise, logs the output, unformatted and unstripped, as DEBUG

        See Also:
            ``subprocess.check_output``, which only returns stdout
        """
        log.debug("Calling '{}'".format(" ".join(cmd)))
        kwargs = copy(kwargs)
        if "cwd" in kwargs and isinstance(kwargs["path"], PurePath):
            kwargs["cwd"] = str(kwargs["cwd"])
        try:
            x = subprocess.run(
                *[str(c) for c in cmd],
                capture_output=True,
                check=True,
                text=True,
                encoding="utf8",
                **kwargs,
            )
            log.debug(f"stdout: '{x.stdout}'")
            log.debug(f"stderr: '{x.stderr}'")
            x.stdout = x.stdout.strip()
            x.stderr = x.stderr.strip()
            return x
        except subprocess.CalledProcessError as e:
            cls.log_called_process_error(e, log_fn=log.error)
            raise

    @classmethod
    def log_called_process_error(
        cls,
        e: subprocess.CalledProcessError,
        *,
        log_fn: Callable[[str], None],
    ) -> None:
        """
        Outputs some formatted text describing the error with its full stdout and stderr.

        Args:
            e: The error
            log_fn: For example, ``logger.warning``
        """
        log_fn(f'Failed on command: {" ".join(e.cmd)}')
        out = None
        if e.stdout is not None:
            out = e.stdout.decode(encoding="utf8") if isinstance(e.stdout, bytes) else e.stdout
        log_fn("《no stdout》" if out is None else f"stdout: {out}")
        err = None
        if e.stderr is not None:
            err = e.stderr.decode(encoding="utf8") if isinstance(e.stderr, bytes) else e.stderr
        log_fn("《no stderr》" if err is None else f"stderr: {err}")

    @classmethod
    def stream_cmd_call(
        cls,
        cmd: Sequence[str],
        *,
        callback: Callable[[bool, bytes], None] = None,
        timeout_secs: Optional[float] = None,
        **kwargs,
    ) -> None:
        """
        Processes stdout and stderr on separate threads.
        Streamed -- can avoid filling a stdout or stderr buffer.
        Calls an external command, waits, and throws a
        ExternalCommandFailed for nonzero exit codes.

        Args:
            cmd: The command args
            callback: A function that processes (is_stderr, piped line).
                      If ``None``, uses :meth:`smart_log`.
            timeout_secs: Max seconds to wait for the next line
            kwargs: Passed to ``subprocess.Popen``; do not pass ``stdout`` or ``stderr``.

        Raises:
            CalledProcessError: If the exit code is nonzero
        """
        if callback is None:
            callback = cls.smart_log
        cmd = [str(p) for p in cmd]
        logger.debug("Streaming '{}'".format(" ".join(cmd)))
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
        try:
            q = Queue()
            Thread(target=cls._reader, args=[False, p.stdout, q]).start()
            Thread(target=cls._reader, args=[True, p.stderr, q]).start()
            for _ in range(2):
                for source, line in iter(q.get, None):
                    callback(source, line)
            exit_code = p.wait(timeout=timeout_secs)
        finally:
            p.kill()
        if exit_code != 0:
            raise subprocess.CalledProcessError(
                exit_code, " ".join(cmd), "<<unknown>>", "<<unknown>>"
            )

    @classmethod
    def smart_log(
        cls,
        is_stderr: bool,
        line: bytes,
        *,
        prefix: str = "",
        log: logging.Logger = logger,
    ) -> None:
        """
        Maps (is_stderr, piped line) pairs to logging statements.
        The data must be utf8-encoded.
        If the line starts with ``warning:`` (case-insensitive), uses ``log.warning``.
        The same is true for any other method attached to ``log``.
        Falls back to DEBUG if no valid prefix is found.
        This is useful if you wrote an external application (e.g. in C)
        and want those logging statements mapped into your calling Python code.
        """
        line = line.decode("utf-8")
        try:
            fn = getattr(log, line.split(":")[0].lower())
        except AttributeError:
            fn = log.debug
        source = "stderr" if is_stderr else "stdout"
        fn(f"{fn.__name__.upper()} [{source}]: {line}")
        if line.lower().startswith("FATAL:"):
            logger.fatal(prefix + line)
        elif line.startswith("ERROR:"):
            logger.error(prefix + line)
        elif line.startswith("WARNING:"):
            logger.warning(prefix + line)
        elif line.startswith("INFO:"):
            logger.info(prefix + line)
        elif line.startswith("DEBUG:"):
            logger.debug(prefix + line)
        else:
            logger.debug(prefix + line)

    @classmethod
    def _disp(cls, out, ell, name):
        out = out.strip()
        if "\n" in out:
            ell(name + ":\n<<=====\n" + out + "\n=====>>")
        elif len(out) > 0:
            ell(name + ": <<===== " + out + " =====>>")
        else:
            ell(name + ": <no output>")

    @classmethod
    def _log(cls, out, err, ell):
        cls._disp(out, ell, "stdout")
        cls._disp(err, ell, "stderr")

    @classmethod
    def _reader(cls, pipe_type, pipe, queue):
        try:
            with pipe:
                for line in iter(pipe.readline, b""):
                    queue.put((pipe_type, line))
        finally:
            queue.put(None)

    @classmethod
    def _wrap(cls, s: str, ell: int) -> str:
        wrapped = textwrap.wrap(s.strip(), ell - 4)
        return textwrap.indent(os.linesep.join(wrapped), "    ")


__all__ = ["CallTools"]
